#include "BC.h"

void BC::setOpenmpConfig(int dynamic, int threads) {
  omp_set_dynamic(dynamic);
  omp_set_num_threads(threads);
}

// TODO: make the Pull variant FASTER !!!!!!!!!!!
// TODO: make the merging in both variants faster (use multiple threads)
// TODO: use sth faster than std::vector for the frontier management
// TODO: I removed std::unique and std::sort as they made things much slower.
// Check if removing duplicates with other means helps somehow and by how much.


void BC::initializeBC() {
  
}

void BC::runPushWithOpenMP(TradArrayNoDegGraphR* G_csr, v_id ROOT, v_id** bfs_out) {
  v_id n = G_csr->n_;
  double* bc_score = new double[n]();
#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    bc_score[i] = 0;
  }

  vector<v_id>** P = new vector<v_id>*[n];
  vector<vector<v_id>*>* S = new vector<vector<v_id>*>();
  v_id* sigma = new v_id[n]();
  v_id* d = new v_id[n]();
  v_id phase = 0;
  v_id count = 0;
  double* delta = new double[n]();

  for(v_id s = 0; s < n; ++s) {

#pragma omp parallel for
    for(v_id t = 0; t < n; ++t) {
      P[t] = new vector<v_id>();
      sigma[t] = 0;
      d[t] = -1;
    }
    sigma[s] = 1;
    d[s] = 0;
    phase = 0;
    S->push_back(new vector<v_id>());
    S[phase]->push(s);
    count = 1;

    TTASLock* P_locks = new TTASLock[n];

    while(count > 0) {
      count = 0;

#pragma omp parallel for
      for(v_id i = 0; i < S[phase]->size(); ++i) {
        v_id v = S[phase][i];
        v_id* offsets = G_csr->offsets_;
        v_id* adj = G_csr->adj_;
        v_id v_deg = offsets[v+1] - offsets[v];

#pragma omp parallel for
        for(v_id e = 0; e < v_deg; ++e) {
          v_id w = adj[ offsets[v] + e ];
          v_id dw = __sync_val_compare_and_swap(&d[w], -1, phase + 1);
          if(dw == -1) {
            v_id p = fetch_and_add(&count, 1);
            (S[phase + 1])[p] = w;
            dw = phase + 1;
          }
          if(dw == phase + 1) {
              P_locks[w].lock();
              P[w]->push_back(v);
              P_locks[w].unlock();
          }
        }
      }
      ++phase;
    }
    --phase;

    // Part III: 

#pragma omp parallel for
    for(v_id t = 0; t < n; ++t) {
      delta[t] = 0;
    }

    while(phase > 0) {

#pragma omp parallel for
      for(v_id i = 0; i < S[phase]->size(); ++i) {
        v_id w = S[phase][i];
        
        for(v_id j = 0; j < P[w]->size(); ++j) {
          v_id v = P[w][j];
          v_locks[v].lock();
          delta[v] += ((double)sigma[v] / (double)sigma[w]) * (1 + delta[w]);
          v_locks[v].unlock();
        }

        bc_score[w] += delta[w];
      }
      --phase;
    }
  }



  v_id frontier_size = 1;
  v_id next_frontier_size = 0;
  v_id* frontier = new v_id[frontier_size]();
  v_id* elements_inserted_global = NULL;
  frontier[0] = ROOT;
  parent[ROOT] = ROOT;

#pragma omp parallel 
  {
    v_id k = 0;
    v_id* adj = G_csr->adj_;
    v_id* offsets = G_csr->offsets_;
    int thread_id = omp_get_thread_num();
    int num_threads = omp_get_num_threads();
    int itr = 0;
    v_id next_frontier_part_size = 0;
    v_id* next_frontier_part = NULL;
    v_id elements_inserted = 0;

#pragma omp single
    {
      elements_inserted_global = new v_id[num_threads];
    }

#pragma omp barrier
    next_frontier_part_size = ceil( ((double)n) / (num_threads) );
    next_frontier_part = new v_id[next_frontier_part_size];

    while (frontier_size > 0) {
#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", itr);
      }
#endif

#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

      elements_inserted = 0;
      elements_inserted_global[thread_id] = 0;

#pragma omp for
      //for (k = 0; k < (v_id)frontier.size(); ++k)
      for (k = 0; k < frontier_size; ++k) {
        const v_id v = frontier[k];

        v_id v_offset = offsets[v];
        v_id v_deg = offsets[v+1] - v_offset;
        v_id v_neigh = -1;

        for(v_id id = 0; id < v_deg; ++id) {
          v_neigh = adj[v_offset + id];
          if(parent[v_neigh] == -1) {
            if(__sync_bool_compare_and_swap(&parent[v_neigh], -1, v)) {
              parent[v_neigh] = v; // TODO: try the no-CAS variant
              next_frontier_part[elements_inserted++] = v_neigh;
              if(elements_inserted == next_frontier_part_size) {
                v_id* larger_next_frontier_part = new v_id[ 2*next_frontier_part_size ];
                memcpy(&larger_next_frontier_part[0], &next_frontier_part[0], next_frontier_part_size * sizeof(v_id));
                delete [] next_frontier_part;
                next_frontier_part = larger_next_frontier_part;
                next_frontier_part_size *= 2;
              }
            }
          }
        }
      }

      __sync_fetch_and_add(&next_frontier_size, elements_inserted);
      elements_inserted_global[thread_id] = elements_inserted;

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
        LSB_Rec(run_id);

        LSB_Set_Rparam_string("phs", "merg");
        LSB_Res();
      }
#endif

#pragma omp barrier

#pragma omp single
      {
        __sync_synchronize();
        delete [] frontier; //TODO: we can remove one barrier at the cost of more memory usage
        frontier = new v_id[next_frontier_size];
        frontier_size = next_frontier_size;
        next_frontier_size = 0;
      }

#pragma omp barrier

      v_id my_offset = 0;
      for(int i = 0; i < thread_id; ++i) {
        my_offset += elements_inserted_global[i];
      }

      assert(my_offset < frontier_size || (my_offset == frontier_size && elements_inserted == 0));
      memcpy(&frontier[my_offset], &next_frontier_part[0], elements_inserted * sizeof(v_id));

      ++itr;

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
        LSB_Rec(run_id);
      }
#endif
    }

    delete [] next_frontier_part;
  }

  *bfs_out = parent;

  delete [] elements_inserted_global;
  delete [] frontier;
}

void BC::runPullWithOpenMP(TradArrayNoDegGraphR* G_csr, v_id ROOT, v_id** bfs_out) {
  v_id n = G_csr->n_;

  v_id* parent = new v_id[n]();

#pragma omp parallel for
  for (unsigned i = 0; i < n; i++) {
    parent[i] = -1;
  }

  v_id frontier_size = 1;
  v_id next_frontier_size = 0;
  v_id* frontier = new v_id[frontier_size]();
  v_id* elements_inserted_global = NULL;
  frontier[0] = ROOT;
  parent[ROOT] = ROOT;

  uint64_t* in_frontier = new uint64_t[(v_id)ceil( (1.0 * n) / BITS_IN_64BIT_WORD )]();
  v_id root_bit_base_pos = ROOT / BITS_IN_64BIT_WORD;
  in_frontier[root_bit_base_pos] = in_frontier[root_bit_base_pos] | ( ((uint64_t)1) << (ROOT % BITS_IN_64BIT_WORD));

#pragma omp parallel 
  {
    v_id v = 0;
    v_id k = 0;
    v_id* adj = G_csr->adj_;
    v_id* offsets = G_csr->offsets_;
    int thread_id = omp_get_thread_num();
    int num_threads = omp_get_num_threads();
    int itr = 0;
    v_id next_frontier_part_size = 0;
    v_id* next_frontier_part = NULL;
    v_id elements_inserted = 0;

#pragma omp single
    {
      elements_inserted_global = new v_id[num_threads];
    }

#pragma omp barrier

    next_frontier_part_size = ceil( ((double)n) / (num_threads) );
    next_frontier_part = new v_id[next_frontier_part_size];

    while (frontier_size > 0) {
#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", itr);

        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

      elements_inserted = 0;
      elements_inserted_global[thread_id] = 0;

#pragma omp for
      for (v = 0; v < n; ++v) {
        if(parent[v] == -1) {
          v_id v_offset = offsets[v];
          v_id v_deg = offsets[v+1] - v_offset;
          v_id v_neigh = -1;
          v_id v_neigh_bit_base_pos = -1;
          int if_in_frontier = -1;

          for(v_id id = 0; id < v_deg; ++id) {
            v_neigh = adj[v_offset + id];
            v_neigh_bit_base_pos = v_neigh / BITS_IN_64BIT_WORD;
            if_in_frontier = (in_frontier[v_neigh_bit_base_pos] >> (v_neigh % BITS_IN_64BIT_WORD)) & ((uint64_t)1);
            assert(if_in_frontier == 0 || if_in_frontier == 1);
            if(if_in_frontier == 1) {
              parent[v] = v_neigh; 
              next_frontier_part[elements_inserted++] = v;
              if(elements_inserted == next_frontier_part_size) {
                v_id* larger_next_frontier_part = new v_id[ 2*next_frontier_part_size ];
                memcpy(&larger_next_frontier_part[0], &next_frontier_part[0], next_frontier_part_size * sizeof(v_id));
                delete [] next_frontier_part;
                next_frontier_part = larger_next_frontier_part;
                next_frontier_part_size *= 2;
              }
            }
          }
        }
      }

      __sync_fetch_and_add(&next_frontier_size, elements_inserted);
      elements_inserted_global[thread_id] = elements_inserted;

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
        LSB_Rec(run_id);

        LSB_Set_Rparam_string("phs", "merg");
        LSB_Res();
      }
#endif

#pragma omp barrier

#pragma omp single
      {
        __sync_synchronize();
        delete [] frontier; //TODO: we can remove one barrier at the cost of more memory usage
        frontier = new v_id[next_frontier_size];
        frontier_size = next_frontier_size;
        next_frontier_size = 0;
      }

#pragma omp barrier

      v_id my_offset = 0;
      for(int i = 0; i < thread_id; ++i) {
        my_offset += elements_inserted_global[i];
      }

      assert(my_offset < frontier_size || (my_offset == frontier_size && elements_inserted == 0));
      memcpy(&frontier[my_offset], &next_frontier_part[0], elements_inserted * sizeof(v_id));

#pragma omp barrier

      // Clear the current in_frontier.
      v_id nr = (v_id)ceil( (1.0 * n) / BITS_IN_64BIT_WORD );

#pragma omp for
      for (k = 0; k < nr; ++k) {
        in_frontier[k] = 0;
      }

#pragma omp for
      for (k = 0; k < frontier_size; ++k) {
        v = frontier[k];
        __sync_fetch_and_or(&in_frontier[v / BITS_IN_64BIT_WORD], (uint64_t)1 << (v % BITS_IN_64BIT_WORD) );
      }

      ++itr;

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
        LSB_Rec(run_id);
      }
#endif
    }

    delete [] next_frontier_part;
  }

  *bfs_out = parent;

  delete [] elements_inserted_global;
  delete [] frontier;
  delete [] in_frontier;
}


/*

TODO: fix and use the below probably faster version

#pragma omp parallel shared(k1,k2) 
{
v_id k = 0;
v_id* nbuf = new v_id[THREAD_BUF_LEN];
v_id* adj = G_csr->adj_;
v_id* offsets = G_csr->offsets_;

while (k1 != k2) {
const v_id oldk2 = k2;
v_id kbuf = 0;

#pragma omp barrier
#pragma omp for
for (k = k1; k < oldk2; ++k) {
const v_id v = vlist[k];

v_id v_offset = offsets[v];
v_id* v_neighs = &adj[v_offset];
v_id v_deg = offsets[v+1] - v_offset;
v_id v_neigh = -1;

for(v_id id = 0; id < v_deg; ++id) {
v_neigh = v_neighs[id];
parent[v_neigh] = v; // TODO: try the if-CAS variant
if (kbuf < THREAD_BUF_LEN) {
nbuf[kbuf++] = v_neigh;
} 
else {
v_id voff = __sync_fetch_and_add(&k2, THREAD_BUF_LEN), vk;
assert (voff + THREAD_BUF_LEN <= n);

for (vk = 0; vk < THREAD_BUF_LEN; ++vk) {
vlist[voff + vk] = nbuf[vk];
}

nbuf[0] = v_neigh;
kbuf = 1;
}
}
}
if (kbuf) {
v_id voff = __sync_fetch_and_add(&k2, kbuf), vk;
assert (voff + kbuf <= n);
for (vk = 0; vk < kbuf; ++vk) {
vlist[voff + vk] = nbuf[vk];
}
}

#pragma omp single
k1 = oldk2;

#pragma omp barrier
}
}

 */


  int
BC::compute_levels (int64_t * level,
    int64_t nv, const int64_t *  bfs_tree, int64_t root)
{
  int err = 0;

#pragma omp parallel shared(err) 
  {
    int terr;
    int64_t k;

#pragma omp for
    for (k = 0; k < nv; ++k)
      level[k] = (k == root? 0 : -1);

#pragma omp for 
    for (k = 0; k < nv; ++k) {
      if (level[k] >= 0) continue;
      terr = err;
      if (!terr && bfs_tree[k] >= 0 && k != root) {
        int64_t parent = k;
        int64_t nhop = 0;
        /* Run up the tree until we encounter an already-leveled vertex. */
        while (parent >= 0 && level[parent] < 0 && nhop < nv) {
          assert (parent != bfs_tree[parent]);
          parent = bfs_tree[parent];
          ++nhop;
        }
        if (nhop >= nv) terr = -1; /* Cycle. */
        if (parent < 0) terr = -2; /* Ran off the end. */

        if (!terr) {
          /* Now assign levels until we meet an already-leveled vertex */
          /* NOTE: This permits benign races if parallelized. */
          nhop += level[parent];
          parent = k;
          while (level[parent] < 0) {
            assert (nhop > 0);
            level[parent] = nhop--;
            parent = bfs_tree[parent];
          }
          assert (nhop == level[parent]);

          /* Internal check to catch mistakes in races... */
#if !defined(NDEBUG)
          nhop = 0;
          parent = k;
          int64_t lastlvl = level[k]+1;
          while (level[parent] > 0) {
            assert (lastlvl == 1 + level[parent]);
            lastlvl = level[parent];
            parent = bfs_tree[parent];
            ++nhop;
          }
#endif
        }
      }
      if (terr) { 
        err = terr; 
#pragma omp flush (err) 
      }
    }
  }
  return err;
}

  int64_t
BC::verify_bfs_tree (int64_t *bfs_tree_in, int64_t max_bfsvtx,
    int64_t root,
    const int64_t *IJ_in, int64_t nedge)
{
  int64_t *  bfs_tree = bfs_tree_in;
  const int64_t *  IJ = IJ_in;

  int err, nedge_traversed;
  int64_t *  seen_edge, *  level;

  const int64_t nv = max_bfsvtx+1;

  /*
     This code is horrifically contorted because many compilers
     complain about continue, return, etc. in parallel sections.
   */

  if (root > max_bfsvtx || bfs_tree[root] != root)
    return -999;

  err = 0;
  nedge_traversed = 0;
  seen_edge = (int64_t*) malloc (2 * (nv) * sizeof (*seen_edge));
  level = &seen_edge[nv];

  err = compute_levels (level, nv, bfs_tree, root);

  if (err) goto done;

#pragma omp parallel shared(err)
  {
    int64_t k;
    int terr = 0;
#pragma omp for 
    for (k = 0; k < nv; ++k)
      seen_edge[k] = 0;

#pragma omp for reduction(+:nedge_traversed) 
    for (k = 0; k < 2*nedge; k+=2) {
      const int64_t i = IJ[k];
      const int64_t j = IJ[k+1];
      int64_t lvldiff;
      terr = err;

      if (i < 0 || j < 0) continue;
      if (i > max_bfsvtx && j <= max_bfsvtx) terr = -10;
      if (j > max_bfsvtx && i <= max_bfsvtx) terr = -11;
      if (terr) { 
        err = terr; 
#pragma omp flush(err) 
      }
      if (terr || i > max_bfsvtx /* both i & j are on the same side of max_bfsvtx */)
        continue;

      /* All neighbors must be in the tree. */
      if (bfs_tree[i] >= 0 && bfs_tree[j] < 0) terr = -12;
      if (bfs_tree[j] >= 0 && bfs_tree[i] < 0) terr = -13;
      if (terr) { 
        err = terr; 
#pragma omp flush(err)
      }
      if (terr || bfs_tree[i] < 0 /* both i & j have the same sign */)
        continue;

      /* Both i and j are in the tree, count as a traversed edge.

NOTE: This counts self-edges and repeated edges.  They're
part of the input data.
       */
      ++nedge_traversed;
      /* Mark seen tree edges. */
      if (i != j) {
        if (bfs_tree[i] == j)
          seen_edge[i] = 1;
        if (bfs_tree[j] == i)
          seen_edge[j] = 1;
      }
      lvldiff = level[i] - level[j];
      /* Check that the levels differ by no more than one. */
      if (lvldiff > 1 || lvldiff < -1)
        terr = -14;
      if (terr) { 
        err = terr; 
#pragma omp flush(err) 
      }
    }

    if (!terr) {
      /* Check that every BC edge was seen and that there's only one root. */
#pragma omp for  
      for (k = 0; k < nv; ++k) {
        terr = err;
        if (!terr && k != root) {
          if (bfs_tree[k] >= 0 && !seen_edge[k])
            terr = -15;
          if (bfs_tree[k] == k)
            terr = -16;
          if (terr) { 
            err = terr; 
#pragma omp flush(err)
          }
        }
      }
    }
  }
done:

  free (seen_edge);
  if (err) return err;
  return nedge_traversed;
}

