// Improved version of BGC (+ end condition) - think about it

#include "BGC.h"
#include <array>
#include "tbb/tbb.h"

void BGC::setOpenmpConfig(int dynamic, int threads) {
  omp_set_dynamic(dynamic);
  omp_set_num_threads(threads);
}

void BGC::runPullWithOpenMP(TradArrayLocalityAwareNoDegGraphR* G_csr, int maxiter, int limit, int** bgc_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  v_id n = 0;
  int* bgc_vals = NULL;
  int** available = NULL;

#pragma omp parallel
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

#ifdef __BENCHMARK__
#pragma omp single
    {
      LSB_Set_Rparam_int("itr", 0);
      LSB_Set_Rparam_string("phs", "prep");
      LSB_Res();
    }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.start_papi_daint_counters();
#endif

#pragma omp single
    {
      n = G_csr->n_;
      bgc_vals = new int[n]();
      available = new int*[n];
    }

#pragma omp for
    for (int i = 0; i < n; ++i) {
      available[i] = new int[limit]();
    }

    int iter = 1;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.stop_papi_daint_counters();
    PapiStats::increase_global_stats(&papi_s_global, &papi_s);
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
    {
#ifdef __XGRAPH_COUNT_OPERATIONS__
      s_global.record_lock_stats_in_lsb();
      papi_s_global.record_papi_in_lsb();
#endif

      LSB_Rec(run_id);
    }
#endif

#pragma omp barrier

    bool check = true;

    while (check) {
#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "phs_1");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      int thread_id = PAPI_thread_id();
#else
      int thread_id = omp_get_thread_num();
#endif

      check = false;

      v_id* my_adj_local = (G_csr->adj_local_)[thread_id];
      v_id* my_offsets_local = (G_csr->offsets_local_)[thread_id];

      v_id my_n_local = (G_csr->n_local_)[thread_id];

      for (v_id id = 0; id < my_n_local; ++id) {

        v_id id_global = G_csr->getVertexGlobalID(id, thread_id);
        v_id offset_local = my_offsets_local[id];
        v_id deg_local = my_offsets_local[id+1] - my_offsets_local[id];

        bool change = false;

        for (v_id i = 0; i < deg_local; ++i) {
          if (available[id_global][bgc_vals[my_adj_local[offset_local + i]]] == 0 && bgc_vals[my_adj_local[offset_local + i]] >= 0 && bgc_vals[my_adj_local[offset_local + i]] < limit){
            ++available[id_global][bgc_vals[my_adj_local[offset_local + i]]];
            change = true;
          }
        }

        if (iter == 1 || change) {
            int cr = 1;
            for (; cr < limit; cr++)
              if (available[id_global][cr] == 0) break;

            bgc_vals[id_global] = cr;
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);

        LSB_Set_Rparam_string("phs", "phs_2");
        LSB_Res();
      }
#endif

#pragma omp barrier

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

      v_id* my_adj_remote = (G_csr->adj_remote_)[thread_id];

      my_offsets_local = (G_csr->offsets_local_)[thread_id];
      v_id* my_offsets_remote = (G_csr->offsets_remote_)[thread_id];

      my_n_local = (G_csr->n_local_)[thread_id];

      for (v_id id = 0; id < my_n_local; ++id) {

        v_id id_global = G_csr->getVertexGlobalID(id, thread_id);

        v_id offset_remote = my_offsets_remote[id];

        v_id deg_local = my_offsets_local[id+1] - my_offsets_local[id];
        v_id deg_remote = my_offsets_remote[id+1] - my_offsets_remote[id];

        bool change = false;

        if (deg_local < deg_remote) {
          for (v_id i = 0; i < deg_remote; ++i) {
              if (bgc_vals[my_adj_remote[offset_remote + i]] == bgc_vals[id_global]
                  && available[my_adj_remote[offset_remote + i]][bgc_vals[id_global]] == 0
                  && bgc_vals[my_adj_remote[offset_remote + i]] < limit){
                __sync_fetch_and_add(&available[id_global][bgc_vals[my_adj_remote[offset_remote + i]]], 1);
                check = change = true;
              }
          }

#ifdef __XGRAPH_COUNT_OPERATIONS__
          s.omp_critical_ += deg_remote;
#endif
        }

        if (change) {
            int cr = 1;
            for (; cr < limit; cr++)
              if (available[id_global][cr] == 0) break;

            bgc_vals[id_global] = cr;
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0; 
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

      ++iter;

#pragma omp barrier
    }
  }

  *bgc_out = bgc_vals;
}

void BGC::runPushWithOpenMP(TradArrayLocalityAwareNoDegGraphR* G_csr, int maxiter, int limit, int** bgc_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  int n = G_csr->n_;
  int* bgc_vals = new int[n];
  int** available = new int*[n]();

  for (int i = 0; i < n; ++i) {
    bgc_vals[i] = -1;
    available[i] = new int[limit]();
    //for (int it = 0; it < limit; ++it) {
    //  available[i][it] = 0;
    //}
  }

#pragma omp parallel
  { 
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    int iter = 1;

#pragma omp barrier

    bool check = true;

    while (check) {
#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "phs_1");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      int thread_id = PAPI_thread_id();
#else
      int thread_id = omp_get_thread_num();
#endif

      check = false;

      v_id* my_adj_local = (G_csr->adj_local_)[thread_id];
      v_id* my_offsets_local = (G_csr->offsets_local_)[thread_id];

      v_id my_n_local = (G_csr->n_local_)[thread_id];

      for (v_id id = 0; id < my_n_local; ++id) {

        v_id id_global = G_csr->getVertexGlobalID(id, thread_id);
        v_id offset_local = my_offsets_local[id];
        v_id deg_local = my_offsets_local[id+1] - my_offsets_local[id];

        for (v_id i = 0; i < deg_local; ++i) {
          if (bgc_vals[my_adj_local[offset_local + i]] >= 0 && bgc_vals[my_adj_local[offset_local + i]] < limit && available[id_global][bgc_vals[my_adj_local[offset_local + i]]] == 0){
            ++available[id_global][bgc_vals[my_adj_local[offset_local + i]]];
          }
        }

        int cr = 0;
        for (; cr < limit; cr++)
          if (available[id_global][cr] == 0) break;

        bgc_vals[id_global] = cr;
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);

        LSB_Set_Rparam_string("phs", "phs_2");
        LSB_Res();
      }
#endif

#pragma omp barrier

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

      v_id* my_adj_remote = (G_csr->adj_remote_)[thread_id];

      my_offsets_local = (G_csr->offsets_local_)[thread_id];
      v_id* my_offsets_remote = (G_csr->offsets_remote_)[thread_id];

      my_n_local = (G_csr->n_local_)[thread_id];

      for (v_id id = 0; id < my_n_local; ++id) {

        v_id id_global = G_csr->getVertexGlobalID(id, thread_id);

        v_id offset_remote = my_offsets_remote[id];

        v_id deg_local = my_offsets_local[id+1] - my_offsets_local[id];
        v_id deg_remote = my_offsets_remote[id+1] - my_offsets_remote[id];

        if (deg_local < deg_remote) {
          for (v_id i = 0; i < deg_remote; ++i) {
//#pragma omp critical
//            {
              if (bgc_vals[my_adj_remote[offset_remote + i]] == bgc_vals[id_global]
                  && available[id_global][bgc_vals[my_adj_remote[offset_remote + i]]] == 0
                  && bgc_vals[id_global] < limit){
                  __sync_fetch_and_add(&available[my_adj_remote[offset_remote + i]][bgc_vals[id_global]], 1);
                  check = true;
//                ++available[my_adj_remote[offset_remote + i]][bgc_vals[id_global]];
              }
//            }
          }

#ifdef __XGRAPH_COUNT_OPERATIONS__
          s.omp_critical_ += deg_remote;
#endif
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

      ++iter;

#pragma omp barrier

    }
  }


  *bgc_out = bgc_vals;
}

void BGC::runGreedyColoring(TradArrayNoDegGraphR* G_csr, int** bgc_out) {
  v_id n = G_csr->n_;

  int* bgc_vals = new int[n];
  int* available = new int[n];

  for (unsigned i = 0; i < n; ++i) {
    bgc_vals[i] = -1;
  }

  for (v_id id = 0; id < n; ++id) {

    v_id* offsets = G_csr->offsets_;
    v_id* adj = G_csr->adj_;

    for (v_id it = 0; it < n; ++it) {
      available[it] = true;
    }

    v_id id_offset = offsets[id];
    v_id neighs_v_nr = offsets[id + 1] - offsets[id];

    for (v_id i = 0; i < neighs_v_nr; ++i) {

      if (bgc_vals[adj[id_offset + i]] >= 0)
        available[bgc_vals[adj[id_offset + i]]] = false;

    }

    int cr;
    for (cr = 0; cr < n; cr++)
      if (available[cr]) break;

    bgc_vals[id] = cr;
  }

  *bgc_out = bgc_vals;

  //  just print results:
  for (unsigned i = 0; i < n; i++) {
    cout << "VERTEX " << i+1 << ": " << bgc_vals[i] << endl;
  }
}

void BGC::newGraphColoring(TradArrayNoDegGraphR* G_csr, int** sp) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  v_id n = G_csr->n_;
  int* bucket = new int[n]();

  v_id* offsets = G_csr->offsets_;
  v_id* adj = G_csr->adj_;

  tbb::concurrent_vector<tbb::concurrent_vector<int>> buckets;
  tbb::concurrent_vector<int> *nbucket = new tbb::concurrent_vector<int>();

  buckets.push_back(*nbucket);

  for (v_id id = 0; id < n; ++id) {

    v_id id_offset = offsets[id];
    v_id neighs_v_nr = offsets[id + 1] - offsets[id];

    bool check = true;
    for (v_id i = 0; i < neighs_v_nr; ++i) {
      if (bucket[adj[id_offset + i]] != 0) {
        check = false;
      }
    }
    if (check) {
      bucket[id] = 1;
      buckets.at(0).push_back(id);
    }
  }

  int iter = 0;

#pragma omp parallel
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (!buckets.at(iter).empty()) {

#pragma omp single
      {

        nbucket = new tbb::concurrent_vector<int>();
        buckets.push_back(*nbucket);
      }

#ifdef __BENCHMARK__
#pragma omp single
    { 
      LSB_Set_Rparam_int("itr", iter);
      LSB_Set_Rparam_string("phs", "comp");
      LSB_Res();
    }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.start_papi_daint_counters();
#endif



#pragma omp for
      for (int j = 0; j < buckets.at(iter).size(); ++j) {

        int i = buckets.at(iter).at(j);
        int wg = iter + 1;

        v_id id_offset = offsets[i];
        v_id neighs_v_nr = offsets[i + 1] - offsets[i];

        for (v_id id = 0; id < neighs_v_nr; ++id) {
          if (bucket[adj[id_offset + id]] == 0 || (bucket[i] == wg && bucket[adj[id_offset + id]] == wg)) {
            bucket[adj[id_offset + id]] = wg + 1;
            buckets.at(wg).push_back(adj[id_offset + id]);
          }
        }
      }


#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);

        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#pragma omp single
      {

        iter++;
      }

    }
  }

    *sp = bucket;
  }

void BGC::graphColoringGreedySwitch(TradArrayNoDegGraphR* G_csr, int switch_iter, int** sp) {

    v_id n = G_csr->n_;
    int* bucket = new int[n]();

    v_id* offsets = G_csr->offsets_;
    v_id* adj = G_csr->adj_;

    tbb::concurrent_vector<tbb::concurrent_vector<int>> buckets;
    tbb::concurrent_vector<int> *nbucket = new tbb::concurrent_vector<int>();

    buckets.push_back(*nbucket);

    for (v_id id = 0; id < n; ++id) {

        v_id id_offset = offsets[id];
        v_id neighs_v_nr = offsets[id + 1] - offsets[id];

        bool check = true;
        for (v_id i = 0; i < neighs_v_nr; ++i) {
            if (bucket[adj[id_offset + i]] != 0) {
                check = false;
            }
        }
        if (check) {
            bucket[id] = 1;
            buckets.at(0).push_back(id);
        }
    }

    int iter = 0;

    while (!buckets.at(iter).empty() && iter < switch_iter) {

        nbucket = new tbb::concurrent_vector<int>();
        buckets.push_back(*nbucket);

#pragma omp parallel for
        for (int j = 0; j < buckets.at(iter).size(); ++j) {

            int i = buckets.at(iter).at(j);
            int wg = iter + 1;

            v_id id_offset = offsets[i];
            v_id neighs_v_nr = offsets[i + 1] - offsets[i];

            for (v_id id = 0; id < neighs_v_nr; ++id) {
                if (bucket[adj[id_offset + id]] == 0 || (bucket[i] == wg && bucket[adj[id_offset + id]] == wg)) {
                    bucket[adj[id_offset + id]] = wg + 1;
                    buckets.at(wg).push_back(adj[id_offset + id]);
                }
            }
        }

        iter++;

    }

    iter++;

    // greedy means pull!
    for (v_id id = 0; id < n; ++id) {
        if (bucket[id] == 0 || bucket[id] == iter) {
            v_id id_offset = offsets[id];
            v_id neighs_v_nr = offsets[id + 1] - offsets[id];

            bool *available = new bool[n]();

            for (v_id i = 0; i < neighs_v_nr; ++i) {
                if (bucket[adj[id_offset + i]] > 0)
                    available[bucket[adj[id_offset + i]]] = true;
            }

            int cr;
            for (cr = 1; cr < n; cr++)
                if (!available[cr]) break;

            bucket[id] = cr;
        }
    }

    *sp = bucket;
}

void BGC::graphColoringSwitch(TradArrayNoDegGraphR* G_csr, int switch_iter, int** sp) {

    v_id n = G_csr->n_;
    int* bucket = new int[n]();

    v_id* offsets = G_csr->offsets_;
    v_id* adj = G_csr->adj_;

    tbb::concurrent_vector<tbb::concurrent_vector<int>> buckets;
    tbb::concurrent_vector<int> *nbucket = new tbb::concurrent_vector<int>();

    buckets.push_back(*nbucket);

    for (v_id id = 0; id < n; ++id) {

        v_id id_offset = offsets[id];
        v_id neighs_v_nr = offsets[id + 1] - offsets[id];

        bool check = true;
        for (v_id i = 0; i < neighs_v_nr; ++i) {
            if (bucket[adj[id_offset + i]] != 0) {
                check = false;
            }
        }
        if (check) {
            bucket[id] = 1;
            buckets.at(0).push_back(id);
        }
    }

    int iter = 0;

    while (!buckets.at(iter).empty() && iter < switch_iter) {

        nbucket = new tbb::concurrent_vector<int>();
        buckets.push_back(*nbucket);

#pragma omp parallel for
        for (int j = 0; j < buckets.at(iter).size(); ++j) {

            int i = buckets.at(iter).at(j);
            int wg = iter + 1;

            v_id id_offset = offsets[i];
            v_id neighs_v_nr = offsets[i + 1] - offsets[i];

            for (v_id id = 0; id < neighs_v_nr; ++id) {
                if (bucket[adj[id_offset + id]] == 0 || (bucket[i] == wg && bucket[adj[id_offset + id]] == wg)) {
                    bucket[adj[id_offset + id]] = wg + 1;
                    buckets.at(wg).push_back(adj[id_offset + id]);
                }
            }
        }

        iter++;

    }

    iter++;

    // greedy means pull!
    for (v_id id = 0; id < n; ++id) {
        if (bucket[id] == 0) {
            v_id id_offset = offsets[id];
            v_id neighs_v_nr = offsets[id + 1] - offsets[id];

            bool *available = new bool[n]();

            for (v_id i = 0; i < neighs_v_nr; ++i) {
                if (bucket[adj[id_offset + i]] > 0)
                    available[bucket[adj[id_offset + i]]] = true;
            }

            int cr;
            for (cr = 1; cr < n; cr++)
                if (!available[cr]) break;

            bucket[id] = cr;
        }
    }

    *sp = bucket;
}
