#include "MST.h"
//#include "tbb/tbb.h"

void MST::setOpenmpConfig(int dynamic, int threads) {
    omp_set_dynamic(dynamic);
    omp_set_num_threads(threads);
}

// We need general rebuild of this algorithm when we
// have lib version with edge identification

void MST::runPullWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int threads, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n]();

    int max_edge_w = 1000;
    int mst_val = 0;

    int threads_no = threads;

    for (v_id id = 0; id < threads_no; ++id) {
        sv_flag[id] = id+1;
    }

    bool main_check = true;

    int* min_edge_v = new int[threads_no]();
    int* min_edge_id = new int[threads_no]();
    int* min_edge_w = new int[threads_no]();
    bool* checker = new bool[threads_no]();

    while (main_check) {

        main_check = false;

#pragma omp parallel for
        for (v_id id = 1; id <= threads_no; ++id) {

//            int thread_id = omp_get_thread_num();
            int thread_id = id-1;

            min_edge_v[thread_id] = 0;
            min_edge_id[thread_id] = 0;
            min_edge_w[thread_id] = max_edge_w;
            checker[thread_id] = false;

            for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == id) {
                    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
                        if (G_csr->w(i, e_id) < min_edge_w[thread_id] && sv_flag[G_csr->e(i, e_id)] != id) {
                            min_edge_v[thread_id] = i;
                            min_edge_id[thread_id] = e_id;
                            min_edge_w[thread_id] = G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                            checker[thread_id] = true;
                        }
                    }
                }
            }
        }

#pragma omp barrier
        for (v_id id = 0; id < threads_no; ++id) {

            int thread_id = id;

            if (checker[thread_id]) {
                int old_flag = sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])];
                int new_flag = sv_flag[min_edge_v[thread_id]];

                if (old_flag != new_flag) {

                    bool rec = false;

                    if (old_flag) {
                        rec = true;
                        for (v_id id = 0; id < n; ++id) {
                            if (sv_flag[id] == old_flag) {
                                sv_flag[id] = new_flag;
                            }
                        }
                    } else {
                        sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])] = new_flag;
                    }

                    if (rec) {
                        for (v_id id = 0; id < n; ++id) {
                            if (rec && sv_flag[id] == 0) {
                                sv_flag[id] = old_flag;
                                rec = false;
                            }
                        }
                    }

                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                }
            }
        }
        int cur_flag = sv_flag[0];

        for (int i = 1; i < n; ++i) {
            if (sv_flag[i] != cur_flag) {
                main_check = true;
                continue;
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPushWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int threads, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n]();

    int max_edge_w = 1000;
    int mst_val = 0;

    int threads_no = threads;

    for (v_id id = 0; id < threads_no; ++id) {
        sv_flag[id] = id+1;
    }

    bool main_check = true;

    int* min_edge_v = new int[threads_no]();
    int* min_edge_id = new int[threads_no]();
    int* min_edge_w = new int[threads_no]();
    bool* checker = new bool[threads_no]();

    while (main_check) {

        main_check = false;

#pragma omp parallel for
        for (v_id id = 1; id <= threads_no; ++id) {

//            int thread_id = omp_get_thread_num();
            int thread_id = id-1;

            min_edge_v[thread_id] = 0;
            min_edge_id[thread_id] = 0;
            min_edge_w[thread_id] = max_edge_w;
            checker[thread_id] = false;

            for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == id) {
                    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
                        if (G_csr->w(i, e_id) < min_edge_w[thread_id] && sv_flag[G_csr->e(i, e_id)] != id) {
                            min_edge_v[thread_id] = i;
                            min_edge_id[thread_id] = e_id;
                            min_edge_w[thread_id] = G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                            checker[thread_id] = true;
                        }
                    }
                }
            }
        }

#pragma omp barrier
        for (v_id id = 0; id < threads_no; ++id) {

            int thread_id = id;

            if (checker[thread_id]) {
                int old_flag = sv_flag[min_edge_v[thread_id]];
                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])];

                if (old_flag != new_flag) {

                    bool rec = false;

                    if (old_flag && new_flag) {
                        rec = true;
                        for (v_id id = 0; id < n; ++id) {
                            if (sv_flag[id] == old_flag) {
                                sv_flag[id] = new_flag;
                            }
                        }
                    } else {
                        sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])] = old_flag;
                    }

                    if (rec) {
                        for (v_id id = 0; id < n; ++id) {
                            if (rec && sv_flag[id] == 0) {
                                sv_flag[id] = old_flag;
                                rec = false;
                            }
                        }
                    }

                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                }
            }
        }
        int cur_flag = sv_flag[0];

        for (int i = 1; i < n; ++i) {
            if (sv_flag[i] != cur_flag) {
                main_check = true;
                continue;
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPullWithOpenMP___MST_bucket(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    vector<int>* buckets[n];
    vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int old_flag = sv_flag[min_edge_v];
                    int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[old_flag]->size(); ++i) {
                            sv_flag[buckets[old_flag]->at(i)] = new_flag;
                        }
                        buckets[new_flag]->insert(buckets[new_flag]->end(), buckets[old_flag]->begin(), buckets[old_flag]->end());
                        buckets[old_flag]->clear();
                        available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), old_flag), available_buckets.end());
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPushWithOpenMP___MST_bucket(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    vector<int>* buckets[n];
    vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int old_flag = sv_flag[min_edge_v];
                    int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
                            sv_flag[buckets[new_flag]->at(i)] = old_flag;
                        }
                        buckets[old_flag]->insert(buckets[old_flag]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
                        buckets[new_flag]->clear();
                        available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), new_flag), available_buckets.end());
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}