#include "MST.h"
//#include "tbb/tbb.h"

void MST::setOpenmpConfig(int dynamic, int threads) {
    omp_set_dynamic(dynamic);
    omp_set_num_threads(threads);
}

/////////// !!!!!!!!!!!!!!!! FROM MST.cpp:

void MST::runPullWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int threads, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n]();

    int max_edge_w = 1000;
    int mst_val = 0;

    int threads_no = threads;

    for (v_id id = 0; id < threads_no; ++id) {
        sv_flag[id] = id+1;
    }

    bool main_check = true;

    int* min_edge_v = new int[threads_no]();
    int* min_edge_id = new int[threads_no]();
    int* min_edge_w = new int[threads_no]();
    bool* checker = new bool[threads_no]();

    while (main_check) {

        main_check = false;

#pragma omp parallel for
        for (v_id id = 1; id <= threads_no; ++id) {

//            int thread_id = omp_get_thread_num();
            int thread_id = id-1;

            min_edge_v[thread_id] = 0;
            min_edge_id[thread_id] = 0;
            min_edge_w[thread_id] = max_edge_w;
            checker[thread_id] = false;

            for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == id) {
                    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
                        if (G_csr->w(i, e_id) < min_edge_w[thread_id] && sv_flag[G_csr->e(i, e_id)] != id) {
                            min_edge_v[thread_id] = i;
                            min_edge_id[thread_id] = e_id;
                            min_edge_w[thread_id] = G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                            checker[thread_id] = true;
                        }
                    }
                }
            }
        }

#pragma omp barrier
        for (v_id id = 0; id < threads_no; ++id) {

            int thread_id = id;

            if (checker[thread_id]) {
                int old_flag = sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])];
                int new_flag = sv_flag[min_edge_v[thread_id]];

                if (old_flag != new_flag) {

                    bool rec = false;

                    if (old_flag) {
                        rec = true;
                        for (v_id id = 0; id < n; ++id) {
                            if (sv_flag[id] == old_flag) {
                                sv_flag[id] = new_flag;
                            }
                        }
                    } else {
                        sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])] = new_flag;
                    }

                    if (rec) {
                        for (v_id id = 0; id < n; ++id) {
                            if (rec && sv_flag[id] == 0) {
                                sv_flag[id] = old_flag;
                                rec = false;
                            }
                        }
                    }

                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                }
            }
        }
        int cur_flag = sv_flag[0];

        for (int i = 1; i < n; ++i) {
            if (sv_flag[i] != cur_flag) {
                main_check = true;
                continue;
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPushWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int threads, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n]();

    int max_edge_w = 1000;
    int mst_val = 0;

    int threads_no = threads;

    for (v_id id = 0; id < threads_no; ++id) {
        sv_flag[id] = id+1;
    }

    bool main_check = true;

    int* min_edge_v = new int[threads_no]();
    int* min_edge_id = new int[threads_no]();
    int* min_edge_w = new int[threads_no]();
    bool* checker = new bool[threads_no]();

    while (main_check) {

        main_check = false;

#pragma omp parallel for
        for (v_id id = 1; id <= threads_no; ++id) {

//            int thread_id = omp_get_thread_num();
            int thread_id = id-1;

            min_edge_v[thread_id] = 0;
            min_edge_id[thread_id] = 0;
            min_edge_w[thread_id] = max_edge_w;
            checker[thread_id] = false;

            for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == id) {
                    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
                        if (G_csr->w(i, e_id) < min_edge_w[thread_id] && sv_flag[G_csr->e(i, e_id)] != id) {
                            min_edge_v[thread_id] = i;
                            min_edge_id[thread_id] = e_id;
                            min_edge_w[thread_id] = G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                            checker[thread_id] = true;
                        }
                    }
                }
            }
        }

#pragma omp barrier
        for (v_id id = 0; id < threads_no; ++id) {

            int thread_id = id;

            if (checker[thread_id]) {
                int old_flag = sv_flag[min_edge_v[thread_id]];
                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])];

                if (old_flag != new_flag) {

                    bool rec = false;

                    if (old_flag && new_flag) {
                        rec = true;
                        for (v_id id = 0; id < n; ++id) {
                            if (sv_flag[id] == old_flag) {
                                sv_flag[id] = new_flag;
                            }
                        }
                    } else {
                        sv_flag[G_csr->e(min_edge_v[thread_id], min_edge_id[thread_id])] = old_flag;
                    }

                    if (rec) {
                        for (v_id id = 0; id < n; ++id) {
                            if (rec && sv_flag[id] == 0) {
                                sv_flag[id] = old_flag;
                                rec = false;
                            }
                        }
                    }

                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id], min_edge_id[thread_id]);
                }
            }
        }
        int cur_flag = sv_flag[0];

        for (int i = 1; i < n; ++i) {
            if (sv_flag[i] != cur_flag) {
                main_check = true;
                continue;
            }
        }
    }

    *path_weight = mst_val;
}



/////////// !!!!!!!!!!!!!!!! FROM MST_bucket.cpp:
// last modification in MST_bucket by nono: over 2 months ago (noted on May 5):

void MST::runPullWithOpenMP___MST_bucket(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    vector<int>* buckets[n];
    vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int old_flag = sv_flag[min_edge_v];
                    int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[old_flag]->size(); ++i) {
                            sv_flag[buckets[old_flag]->at(i)] = new_flag;
                        }
                        buckets[new_flag]->insert(buckets[new_flag]->end(), buckets[old_flag]->begin(), buckets[old_flag]->end());
                        buckets[old_flag]->clear();
                        available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), old_flag), available_buckets.end());
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPushWithOpenMP___MST_bucket(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    vector<int>* buckets[n];
    vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int old_flag = sv_flag[min_edge_v];
                    int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
                            sv_flag[buckets[new_flag]->at(i)] = old_flag;
                        }
                        buckets[old_flag]->insert(buckets[old_flag]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
                        buckets[new_flag]->clear();
                        available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), new_flag), available_buckets.end());
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}

/////////// !!!!!!!!!!!!!!!! FROM MST_concurrency.cpp:
// last modification in MST_concurrency by nono: over 2 months ago (noted on May 5):

void MST::runPullWithOpenMP___MST_concurrency(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    tbb::concurrent_vector<int>* buckets[n];
    tbb::concurrent_vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new tbb::concurrent_vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int old_flag = sv_flag[min_edge_v];
                    int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[old_flag]->size(); ++i) {
                            sv_flag[buckets[old_flag]->at(i)] = new_flag;
                        }
                        std::copy(buckets[old_flag]->begin(), buckets[old_flag]->end(), buckets[new_flag]->grow_by(buckets[old_flag]->size()));
                        buckets[old_flag]->clear();
                        tbb::concurrent_vector<int> tmp_avail;
                        for (unsigned it = 0; it < available_buckets.size(); ++it) {
                            if (available_buckets.at(it) != old_flag) {
                                tmp_avail.push_back(available_buckets.at(it));
                            }
                        }
                        available_buckets = tmp_avail;
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}

void MST::runPushWithOpenMP___MST_concurrency(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

    v_id n = G_csr->n_;

    int* sv_flag = new int[n];

    int max_edge_w = 0;
    int mst_val = 0;

    tbb::concurrent_vector<int>* buckets[n];
    tbb::concurrent_vector<int> available_buckets;

    for (int i = 0; i < n; ++i) {
        sv_flag[i] = i;
        buckets[i] = new tbb::concurrent_vector<int>();
        buckets[i]->push_back(i);
        available_buckets.push_back(i);
        for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            if (G_csr->w(i, e_id) > max_edge_w) {
                max_edge_w = G_csr->w(i, e_id);
            }
        }
    }

    max_edge_w += 1;

    while (available_buckets.size() > 1) {

#pragma omp parallel for
        for (unsigned j = 0; j < available_buckets.size(); ++j) {

            v_id flag = available_buckets.at(j);

            int min_edge_v = 0;
            int min_edge_id = 0;
            int min_edge_w = max_edge_w;
            bool checker = false;

            for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
                for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
                    if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
                        min_edge_v = buckets[flag]->at(i);
                        min_edge_id = e_id;
                        min_edge_w = G_csr->w(min_edge_v, min_edge_id);
                        checker = true;
                    }
                }
            }

            if (checker) {
#pragma omp critical
                {
                    int new_flag = sv_flag[min_edge_v];
                    int old_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

                    if (old_flag != new_flag) {
                        for (unsigned i = 0; i < buckets[old_flag]->size(); ++i) {
                            sv_flag[buckets[old_flag]->at(i)] = new_flag;
                        }
                        std::copy(buckets[old_flag]->begin(), buckets[old_flag]->end(), buckets[new_flag]->grow_by(buckets[old_flag]->size()));
                        buckets[old_flag]->clear();
                        tbb::concurrent_vector<int> tmp_avail;
                        for (unsigned it = 0; it < available_buckets.size(); ++it) {
                            if (available_buckets.at(it) != old_flag) {
                                tmp_avail.push_back(available_buckets.at(it));
                            }
                        }
                        available_buckets = tmp_avail;
                        mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
                    }
                }
            }
        }
    }

    *path_weight = mst_val;
}





/////////// !!!!!!!!!!!!!!!! FROM MST_old.cpp:
// last modification in MST_old by nono: 26 days ago (noted on May 5):


void MST::runPullWithOpenMP___MST_old(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

  v_id n = G_csr->n_;

  int* sv_flag = new int[n];

  int max_edge_w = 0;

  for (int i = 0; i < n; ++i) {
    sv_flag[i] = i;
    for(int e_id = 0; e_id < G_csr->d(i); ++e_id) {
      if (G_csr->w(i,e_id) > max_edge_w) {
        max_edge_w = G_csr->w(i,e_id);
      }
    }
  }

  max_edge_w += 1;

  bool rep;
  int cur_flag;
  int mst_val = 0;
  int iter = 1;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    do {

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "phs_1");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp single
      {
        rep = false;
      }

#pragma omp for
      for (v_id flag = 0; flag < n; ++flag) {

        int min_edge_v = 0;
        int min_edge_id = 0;
        int min_edge_w = max_edge_w;
        bool checker = false;

        for (int i = 0; i < n; ++i) {
          if (sv_flag[i] == flag) {
            for(int e_id = 0; e_id < G_csr->d(i); ++e_id) {
              if (sv_flag[G_csr->e(i,e_id)] != flag && G_csr->w(i,e_id) < min_edge_w) {
                min_edge_v = i;
                min_edge_id = e_id;
                min_edge_w = G_csr->w(min_edge_v,min_edge_id);
                checker = true;
              }
            }
          }
        }

        if (checker) {
#pragma omp critical
          {
            int old_flag = sv_flag[min_edge_v];
            int new_flag = sv_flag[G_csr->e(min_edge_v,min_edge_id)];

            if (old_flag != new_flag) {
              for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == old_flag)
                  sv_flag[i] = new_flag;
              }
              mst_val = mst_val + G_csr->w(min_edge_v,min_edge_id);
            }
          }
#ifdef __XGRAPH_COUNT_OPERATIONS__
          ++s.omp_critical_;
#endif
        }
      }

#pragma omp single
      {
        cur_flag = sv_flag[0];
        for (int i = 1; i < n; ++i) {
          if (sv_flag[i] != cur_flag) {
            rep = true;
            continue;
          }
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

#pragma omp barrier
    } while(rep);
  }

  *path_weight = mst_val;
}

void MST::runPushWithOpenMP___MST_old(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

  v_id n = G_csr->n_;

  int* sv_flag = new int[n];

  int max_edge_w = 0;

  for (int i = 0; i < n; ++i) {
    sv_flag[i] = i;
    for(int e_id = 0; e_id < G_csr->d(i); ++e_id) {
      if (G_csr->w(i,e_id) > max_edge_w) {
        max_edge_w = G_csr->w(i,e_id);
      }
    }
  }

  max_edge_w += 1;

  bool rep;
  int cur_flag;
  int mst_val = 0;
  int iter = 1;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif    

#pragma omp parallel
  {     
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    do {

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "phs_1");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp single
      {
        rep = false;
      }

#pragma omp for
      for (v_id flag = 0; flag < n; ++flag) {

        int min_edge_v = 0;
        int min_edge_id = 0;
        int min_edge_w = max_edge_w;
        bool checker = false;

        for (int i = 0; i < n; ++i) {
          if (sv_flag[i] == flag) {
            for(int e_id = 0; e_id < G_csr->d(i); ++e_id) {
              if (sv_flag[G_csr->e(i,e_id)] != flag && G_csr->w(i,e_id) < min_edge_w) {
                min_edge_v = i;
                min_edge_id = e_id;
                min_edge_w = G_csr->w(min_edge_v,min_edge_id);
                checker = true;
              }
            }
          }
        }

        if (checker) {
#pragma omp critical
          {
            int old_flag = sv_flag[min_edge_v];
            int new_flag = sv_flag[G_csr->e(min_edge_v,min_edge_id)];

            if (old_flag != new_flag) {
              for (int i = 0; i < n; ++i) {
                if (sv_flag[i] == new_flag)
                  sv_flag[i] = old_flag;
              }
              mst_val = mst_val + G_csr->w(min_edge_v,min_edge_id);
            }
          }
#ifdef __XGRAPH_COUNT_OPERATIONS__
          ++s.omp_critical_;
#endif
        }
      }

#pragma omp single
      {
        cur_flag = sv_flag[0];
        for (int i = 1; i < n; ++i) {
          if (sv_flag[i] != cur_flag) {
            rep = true;
            continue;
          }
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

#pragma omp barrier

    } while(rep);
  }

  *path_weight = mst_val;
}














void MST::runPullWithOpenMP_bucket___MST_old(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

  v_id n = G_csr->n_;

  int* sv_flag = new int[n];

  int max_edge_w = 0;
  int mst_val = 0;

  cerr << "PREP" << endl;

  //vector<int>* buckets[n];
  vector<int>** buckets = new vector<int>*[n];
  vector<int> available_buckets;

  for (int i = 0; i < n; ++i) {
    sv_flag[i] = i;
    buckets[i] = new vector<int>();
    buckets[i]->push_back(i);
    available_buckets.push_back(i);
    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
      if (G_csr->w(i, e_id) > max_edge_w) {
        max_edge_w = G_csr->w(i, e_id);
      }
    }
  }

  max_edge_w += 1;

  cerr << "START" << endl;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel
  {
    int iter = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (available_buckets.size() > 1) {
      #pragma omp single
      {
      cout << "Itr: " << iter << endl;
}

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

      vector<int> tmp_buckets = available_buckets;

#pragma omp for
      for (unsigned j = 0; j < tmp_buckets.size(); ++j) {

        v_id flag = tmp_buckets.at(j);

        int min_edge_v = 0;
        int min_edge_id = 0;
        int min_edge_w = max_edge_w;
        bool checker = false;

        for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
          for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
            if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
              min_edge_v = buckets[flag]->at(i);
              min_edge_id = e_id;
              min_edge_w = G_csr->w(min_edge_v, min_edge_id);
              checker = true;
            }
          }
        }

        if (checker) {
#pragma omp critical
          {
            int old_flag = sv_flag[min_edge_v];
            int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

//            if (old_flag != new_flag) {
            if (old_flag != new_flag && buckets[new_flag]->size() > 0 && buckets[old_flag]->size() > 0) {
              for (unsigned i = 0; i < buckets[old_flag]->size(); ++i) {
                sv_flag[buckets[old_flag]->at(i)] = new_flag;
              }
              buckets[new_flag]->insert(buckets[new_flag]->end(), buckets[old_flag]->begin(), buckets[old_flag]->end());
              buckets[old_flag]->clear();
              available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), old_flag), available_buckets.end());
              mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
            }
          }
#ifdef __XGRAPH_COUNT_OPERATIONS__
          ++s.omp_critical_;
#endif
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

      iter++;
#pragma omp barrier
    }
  }

  *path_weight = mst_val;

  delete [] buckets;
}

void MST::runPushWithOpenMP_bucket___MST_old(TradArrayNoDegEWGraphR* G_csr, int* path_weight) {

  v_id n = G_csr->n_;

  int* sv_flag = new int[n];

  int max_edge_w = 0;
  int mst_val = 0;

  vector<int>** buckets = new vector<int>*[n];
  vector<int> available_buckets;

  for (int i = 0; i < n; ++i) {
    sv_flag[i] = i;
    buckets[i] = new vector<int>();
    buckets[i]->push_back(i);
    available_buckets.push_back(i);
    for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
      if (G_csr->w(i, e_id) > max_edge_w) {
        max_edge_w = G_csr->w(i, e_id);
      }
    }
  }

  max_edge_w += 1;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel
  {
    int iter = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (available_buckets.size() > 1) {
      #pragma omp single
      {
      cout << "Itr: " << iter << endl;
}

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

      vector<int> tmp_buckets = available_buckets;

#pragma omp for
      for (unsigned j = 0; j < tmp_buckets.size(); ++j) {

        v_id flag = tmp_buckets.at(j);

        int min_edge_v = 0;
        int min_edge_id = 0;
        int min_edge_w = max_edge_w;
        bool checker = false;

        for (unsigned i = 0; i < buckets[flag]->size(); ++i) {
          for (int e_id = 0; e_id < G_csr->d(buckets[flag]->at(i)); ++e_id) {
            if (sv_flag[G_csr->e(buckets[flag]->at(i), e_id)] != flag && G_csr->w(buckets[flag]->at(i), e_id) < min_edge_w) {
              min_edge_v = buckets[flag]->at(i);
              min_edge_id = e_id;
              min_edge_w = G_csr->w(min_edge_v, min_edge_id);
              checker = true;
            }
          }
        }

        if (checker) {
#pragma omp critical
          {
            int old_flag = sv_flag[min_edge_v];
            int new_flag = sv_flag[G_csr->e(min_edge_v, min_edge_id)];

//            if (old_flag != new_flag) {
            if (old_flag != new_flag && buckets[new_flag]->size() > 0 && buckets[old_flag]->size() > 0) {
              for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
                sv_flag[buckets[new_flag]->at(i)] = old_flag;
              }
              buckets[old_flag]->insert(buckets[old_flag]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
              buckets[new_flag]->clear();
              available_buckets.erase(std::remove(available_buckets.begin(), available_buckets.end(), new_flag), available_buckets.end());
              mst_val = mst_val + G_csr->w(min_edge_v, min_edge_id);
            }
          }
#ifdef __XGRAPH_COUNT_OPERATIONS__
          ++s.omp_critical_;
#endif
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

      iter++;
#pragma omp barrier
    }
  }

  *path_weight = mst_val;

  delete [] buckets;
}

//void MST::runSwitchedWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int threads_n, int switch_iter_n, int max_edge, int* path_weight) {
//
//    v_id n = G_csr->n_;
//
//    int* sv_flag = new int[n];
//
//    int max_edge_w = max_edge;
//    int mst_val = 0;
//
//    int threads_no = threads_n;
//
//    vector<int> small_edges;
//    //vector<int>* buckets[n];
//    vector<int>** buckets = new vector<int>*[n];
//
//    int* min_edge_v = new int[threads_no]();
//    int* min_edge_id = new int[threads_no]();
//    int* min_edge_w = new int[threads_no]();
//    bool* checker = new bool[threads_no]();
//
//    for (int i = 0; i < n; ++i) {
//        if (i < threads_no) {
//            buckets[i] = new vector<int>();
//            buckets[i]->push_back(i);
//            sv_flag[i] = i;
//        } else {
//            buckets[i] = new vector<int>();
//            sv_flag[i] = -1;
//            small_edges.push_back(i);
//        }
//    }
//
//    int switch_iter = switch_iter_n;
//
//    while (small_edges.size() > switch_iter) {
//
//#pragma omp parallel for
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            min_edge_v[thread_id_get] = 0;
//            min_edge_id[thread_id_get] = 0;
//            min_edge_w[thread_id_get] = max_edge_w;
//            checker[thread_id_get] = false;
//
//            if (!buckets[thread_id_get]->empty())
//                for (unsigned i = 0; i < buckets[thread_id_get]->size(); ++i) {
//                    for (int e_id = 0; e_id < G_csr->d(buckets[thread_id_get]->at(i)); ++e_id) {
//                        if (sv_flag[G_csr->e(buckets[thread_id_get]->at(i), e_id)] != thread_id_get && G_csr->w(buckets[thread_id_get]->at(i), e_id) < min_edge_w[thread_id_get]) {
//                            min_edge_v[thread_id_get] = buckets[thread_id_get]->at(i);
//                            min_edge_id[thread_id_get] = e_id;
//                            min_edge_w[thread_id_get] = G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//                            checker[thread_id_get] = true;
//                        }
//                    }
//                }
//        }
//
//#pragma omp barrier
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            if (checker[thread_id_get]) {
//
//                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])];
//                int j = sv_flag[min_edge_v[thread_id_get]];
//
//                if (j != new_flag) {
//
//                    if (new_flag == -1) {
//                        buckets[j]->push_back(G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get]));
//                        sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])] = j;
//                        small_edges.erase(std::remove(small_edges.begin(), small_edges.end(), G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])), small_edges.end());
//                        small_edges.shrink_to_fit();
//                    } else {
//                        for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
//                            sv_flag[buckets[new_flag]->at(i)] = j;
//                        }
//
//                        buckets[j]->insert(buckets[j]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
//                        buckets[new_flag]->clear();
//
//                        if (!small_edges.empty()) {
//                            buckets[new_flag]->push_back(small_edges.back());
//                            sv_flag[buckets[new_flag]->at(0)] = new_flag;
//                            small_edges.pop_back();
//                            small_edges.shrink_to_fit();
//                        }
//                    }
//                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//                }
//            }
//        }
//    }
//
//    bool finish = false;
//    vector<int> to_remove;
//
//    while (small_edges.size() > 0 && !finish) {
//
//        to_remove.clear();
//
//#pragma omp parallel for
//        for (unsigned j = 0; j < small_edges.size(); ++j) {
//
//            int light_edge = max_edge_w;
//            int min_edge_vs = small_edges.at(j);
//
//            for (int e_id = 0; e_id < G_csr->d(min_edge_vs); ++e_id) {
//                if (G_csr->w(min_edge_vs, e_id) < light_edge) {
//                    light_edge = G_csr->w(min_edge_vs, e_id);
//                }
//            }
//
//            for (int e_id = 0; e_id < G_csr->d(min_edge_vs); ++e_id) {
//                if (sv_flag[min_edge_vs] == -1 && sv_flag[G_csr->e(min_edge_vs, e_id)] != -1 && G_csr->w(min_edge_vs, e_id) == light_edge) {
//                    int old_flag = sv_flag[min_edge_vs];
//                    int new_flag = sv_flag[G_csr->e(min_edge_vs, e_id)];
//
//                    if (old_flag != new_flag) {
//                        sv_flag[min_edge_vs] = new_flag;
//                        buckets[new_flag]->push_back(min_edge_vs);
//                        to_remove.push_back(min_edge_vs);
//                        mst_val = mst_val + G_csr->w(min_edge_vs, e_id);
//                    }
//                }
//            }
//        }
//
//        if (to_remove.size() > 0) {
//            for (unsigned j = 0; j < to_remove.size(); ++j) {
//                small_edges.erase(std::remove(small_edges.begin(), small_edges.end(), to_remove.at(j)), small_edges.end());
//            }
//        } else {
//            finish = true;
//        }
//
//    }
//
//    finish = false;
//
//    while (!finish) {
//
//        #pragma omp parallel for
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            min_edge_v[thread_id_get] = 0;
//            min_edge_id[thread_id_get] = 0;
//            min_edge_w[thread_id_get] = max_edge_w;
//            checker[thread_id_get] = false;
//
//            if (!buckets[thread_id_get]->empty())
//                for (unsigned i = 0; i < buckets[thread_id_get]->size(); ++i) {
//                    for (int e_id = 0; e_id < G_csr->d(buckets[thread_id_get]->at(i)); ++e_id) {
//                        if (sv_flag[G_csr->e(buckets[thread_id_get]->at(i), e_id)] != thread_id_get && G_csr->w(buckets[thread_id_get]->at(i), e_id) < min_edge_w[thread_id_get]) {
//                            min_edge_v[thread_id_get] = buckets[thread_id_get]->at(i);
//                            min_edge_id[thread_id_get] = e_id;
//                            min_edge_w[thread_id_get] = G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//                            checker[thread_id_get] = 1;
//                        }
//                    }
//                }
//        }
//
//        #pragma omp barrier
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            if (checker[thread_id_get]) {
//
//                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])];
//                int j = sv_flag[min_edge_v[thread_id_get]];
//
//                if (j != new_flag) {
//
//                    for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
//                        sv_flag[buckets[new_flag]->at(i)] = j;
//                    }
//
//                    buckets[j]->insert(buckets[j]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
//                    buckets[new_flag]->clear();
//
//                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//
//                    if (buckets[j]->size() == n) finish = true;
//                }
//            }
//        }
//    }
//    *path_weight = mst_val;
//}

void MST::runSwitchedWithOpenMP___MST_old(TradArrayNoDegEWGraphR* G_csr, int threads_n, int switch_iter_n, int max_edge, int* path_weight) {

    v_id n = G_csr->n_;
    
    int* sv_flag = new int[n];

    int max_edge_w = max_edge;
    int mst_val = 0;

    int threads_no = threads_n;

    vector<int> small_edges;
    //vector<int>* buckets[n];
    vector<int>** buckets = new vector<int>*[n];

    int* min_edge_v = new int[threads_no]();
    int* min_edge_id = new int[threads_no]();
    int* min_edge_w = new int[threads_no]();
    bool* checker = new bool[threads_no]();

    for (int i = 0; i < n; ++i) {
        if (i < threads_no) {
            buckets[i] = new vector<int>();
            buckets[i]->push_back(i);
            sv_flag[i] = i;
        } else {
            buckets[i] = new vector<int>();
            sv_flag[i] = -1;
            small_edges.push_back(i);
        }
    }

    int switch_iter = switch_iter_n;
    
//    while (small_edges.size() > switch_iter) {
    for (int ite; ite < 100; ++ite) {
        
#pragma omp parallel for
        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {

            min_edge_v[thread_id_get] = 0;
            min_edge_id[thread_id_get] = 0;
            min_edge_w[thread_id_get] = max_edge_w;
            checker[thread_id_get] = false;

            if (!buckets[thread_id_get]->empty())
                for (unsigned i = 0; i < buckets[thread_id_get]->size(); ++i) {
                    for (int e_id = 0; e_id < G_csr->d(buckets[thread_id_get]->at(i)); ++e_id) {
                        if (sv_flag[G_csr->e(buckets[thread_id_get]->at(i), e_id)] != thread_id_get && G_csr->w(buckets[thread_id_get]->at(i), e_id) < min_edge_w[thread_id_get]) {
                            min_edge_v[thread_id_get] = buckets[thread_id_get]->at(i);
                            min_edge_id[thread_id_get] = e_id;
                            min_edge_w[thread_id_get] = G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
                            checker[thread_id_get] = true;
                        }
                    }
                }
        }

#pragma omp barrier
        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {

            if (checker[thread_id_get]) {

                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])];
                int j = sv_flag[min_edge_v[thread_id_get]];

                if (j != new_flag) {

                    if (new_flag == -1) {
                        buckets[j]->push_back(G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get]));
                        sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])] = j;
                        small_edges.erase(std::remove(small_edges.begin(), small_edges.end(), G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])), small_edges.end());
                        small_edges.shrink_to_fit();
                    } else {
                        for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
                            sv_flag[buckets[new_flag]->at(i)] = j;
                        }

                        buckets[j]->insert(buckets[j]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
                        buckets[new_flag]->clear();

                        if (!small_edges.empty()) {
                            buckets[new_flag]->push_back(small_edges.back());
                            sv_flag[buckets[new_flag]->at(0)] = new_flag;
                            small_edges.pop_back();
                            small_edges.shrink_to_fit();
                        }
                    }
                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
                }
            }
        }
    }

//    bool finish = false;
//    vector<int> to_remove;
//
//    while (small_edges.size() > 0 && !finish) {
//
//        to_remove.clear();
//
//#pragma omp parallel for
//        for (unsigned j = 0; j < small_edges.size(); ++j) {
//
//            int light_edge = max_edge_w;
//            int min_edge_vs = small_edges.at(j);
//
//            for (int e_id = 0; e_id < G_csr->d(min_edge_vs); ++e_id) {
//                if (G_csr->w(min_edge_vs, e_id) < light_edge) {
//                    light_edge = G_csr->w(min_edge_vs, e_id);
//                }
//            }
//
//            for (int e_id = 0; e_id < G_csr->d(min_edge_vs); ++e_id) {
//                if (sv_flag[min_edge_vs] == -1 && sv_flag[G_csr->e(min_edge_vs, e_id)] != -1 && G_csr->w(min_edge_vs, e_id) == light_edge) {
//                    int old_flag = sv_flag[min_edge_vs];
//                    int new_flag = sv_flag[G_csr->e(min_edge_vs, e_id)];
//
//                    if (old_flag != new_flag) {
//                        sv_flag[min_edge_vs] = new_flag;
//                        buckets[new_flag]->push_back(min_edge_vs);
//                        to_remove.push_back(min_edge_vs);
//                        mst_val = mst_val + G_csr->w(min_edge_vs, e_id);
//                    }
//                }
//            }
//        }
//
//        if (to_remove.size() > 0) {
//            for (unsigned j = 0; j < to_remove.size(); ++j) {
//                small_edges.erase(std::remove(small_edges.begin(), small_edges.end(), to_remove.at(j)), small_edges.end());
//            }
//        } else {
//            finish = true;
//        }
//
//    }
//
//    finish = false;
//
//    while (!finish) {
//
//        #pragma omp parallel for
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            min_edge_v[thread_id_get] = 0;
//            min_edge_id[thread_id_get] = 0;
//            min_edge_w[thread_id_get] = max_edge_w;
//            checker[thread_id_get] = false;
//
//            if (!buckets[thread_id_get]->empty())
//                for (unsigned i = 0; i < buckets[thread_id_get]->size(); ++i) {
//                    for (int e_id = 0; e_id < G_csr->d(buckets[thread_id_get]->at(i)); ++e_id) {
//                        if (sv_flag[G_csr->e(buckets[thread_id_get]->at(i), e_id)] != thread_id_get && G_csr->w(buckets[thread_id_get]->at(i), e_id) < min_edge_w[thread_id_get]) {
//                            min_edge_v[thread_id_get] = buckets[thread_id_get]->at(i);
//                            min_edge_id[thread_id_get] = e_id;
//                            min_edge_w[thread_id_get] = G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//                            checker[thread_id_get] = 1;
//                        }
//                    }
//                }
//        }
//
//        #pragma omp barrier
//        for (int thread_id_get = 0; thread_id_get < threads_no; ++thread_id_get) {
//
//            if (checker[thread_id_get]) {
//
//                int new_flag = sv_flag[G_csr->e(min_edge_v[thread_id_get], min_edge_id[thread_id_get])];
//                int j = sv_flag[min_edge_v[thread_id_get]];
//
//                if (j != new_flag) {
//
//                    for (unsigned i = 0; i < buckets[new_flag]->size(); ++i) {
//                        sv_flag[buckets[new_flag]->at(i)] = j;
//                    }
//
//                    buckets[j]->insert(buckets[j]->end(), buckets[new_flag]->begin(), buckets[new_flag]->end());
//                    buckets[new_flag]->clear();
//
//                    mst_val = mst_val + G_csr->w(min_edge_v[thread_id_get], min_edge_id[thread_id_get]);
//
//                    if (buckets[j]->size() == n) finish = true;
//                }
//            }
//        }
//    }
    *path_weight = mst_val;
}
