#include "PR.h"

void PR::setOpenmpConfig(int dynamic, int threads) {
  omp_set_dynamic(dynamic);
  omp_set_num_threads(threads);
}

//TODO: I assume PAPI is initialized!!!

void PR::runPushWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out) {
  v_id n = G_csr->n_;

  double initial = 1.0 / n;
  double* pr_vals = new double[n];
  double* prev_pr_vals = new double[n];
  double random_jump = (1.0 - dump) / n;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    pr_vals[i] = initial;
    prev_pr_vals[i] = 0.0;
  }

  int iter = 1;
  double sum = 0.0;
  double err = 0.0;

#pragma omp parallel 
  {
    v_id* offsets = G_csr->offsets_;
    v_id* adj = G_csr->adj_;

    v_id id_offset = 0; 
    v_id neighs_v_nr = 0;
    double update = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (iter < maxiter) {
#pragma omp single
      {
        for(int i = 0; i < 10; ++i) {
          cout << pr_vals[i] << " ";
        }
        cout << endl;
      }

#pragma omp for
      for (v_id i = 0; i < n; i++) {
        prev_pr_vals[i] = pr_vals[i];
        pr_vals[i] = 0.0;
      }

#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp for
      for (v_id id = 0; id < n; id++) {
        id_offset = offsets[id];
        neighs_v_nr = offsets[id+1] - offsets[id];
        update = dump * prev_pr_vals[id] / neighs_v_nr;

        for(v_id e = 0; e < neighs_v_nr; ++e) {
#pragma omp atomic
          pr_vals[adj[id_offset + e]] += update;
        }

#pragma omp atomic
        pr_vals[id] += random_jump;

#ifdef __XGRAPH_COUNT_OPERATIONS__
        //s.increase_atomic_stats(16 + neighs_v_nr*7, 9 + neighs_v_nr, 0, 0, neighs_v_nr);
        s.increase_atomic_stats(0, 0, neighs_v_nr);
#endif
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_atomic_global_stats(&s_global, &s);
      s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_atomic_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
        //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
      }
#endif

#pragma omp barrier

#pragma omp single
      {
        sum = 0;
      }   

#pragma omp for reduction(+:sum)
      for (v_id i = 0; i < n; i++)
        sum += pr_vals[i];

#pragma omp single
      {
        sum = 1.0 / sum;
      }   

#pragma omp for 
      for (v_id i = 0; i < n; i++)
        pr_vals[i] *= sum;

#pragma omp single
      {
        err = 0;
      }   

#pragma omp for reduction(+:err)
      for (v_id i = 0; i < n; i++)
        err += fabs(pr_vals[i] - prev_pr_vals[i]);

      if (err < tolerance)
        break;

#pragma omp single
      { 
        ++iter;
      }
    }
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;
}





void PR::runPushConvWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double tolerance_v_conv, double** pr_vals_out) {
  v_id n = G_csr->n_;

  double initial = 1.0 / n;
  double* pr_vals = new double[n];
  double* prev_pr_vals = new double[n];
  double random_jump = (1.0 - dump) / n;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    pr_vals[i] = initial;
    prev_pr_vals[i] = 0.0;
  }

  vector<v_id>* adjs = new vector<v_id>[n];
  vector<v_id>* new_adjs = new vector<v_id>[n];
  v_id* G_adj = G_csr->adj_;
  v_id* G_off = G_csr->offsets_;
  vector<v_id> to_remove;

#pragma omp parallel for
  for(v_id v = 0; v < n; ++v) {
    v_id offset = G_off[v];
    v_id deg = G_off[v+1] - G_off[v];
    for(v_id w = 0; w < deg; ++w) {
      adjs[v].push_back(G_adj[offset + w]);
    }
  }

  int iter = 1;
  double sum = 0.0;
  double err = 0.0;

  std::vector<bool> converged;
  for(v_id i = 0; i < n; ++i) {
    converged.push_back(false);
  }

//  int* _converged = new int[n]();

#pragma omp parallel 
  {
    //v_id* offsets = G_csr->offsets_;
    //v_id* adj = G_csr->adj_;

    //v_id id_offset = 0; 
    v_id neighs_v_nr = 0;
    double update = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (iter < maxiter) {
#pragma omp single
      {
        for(int i = 0; i < 10; ++i) {
          cout << pr_vals[i] << " ";
        }
        cout << endl;
      }

#pragma omp for
      for (v_id i = 0; i < n; i++) {
        prev_pr_vals[i] = pr_vals[i];
        pr_vals[i] = 0.0;
      }

#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

/*
#pragma omp for
      for (v_id id = 0; id < n; id++) {
        id_offset = offsets[id];
        neighs_v_nr = offsets[id+1] - offsets[id];
        update = dump * prev_pr_vals[id] / neighs_v_nr;

        for(v_id e = 0; e < neighs_v_nr; ++e) {
          v_id  neigh = adj[id_offset + e];
          //if(!_converged[neigh])
          if(!converged[neigh]) {
#pragma omp atomic
            pr_vals[neigh] += update;

#ifdef __XGRAPH_COUNT_OPERATIONS__
            ++s.fad_;
#endif
          }
        }
*/

#pragma omp for
      for (v_id id = 0; id < n; id++) {
//        neighs_v_nr = offsets[id+1] - offsets[id];
        neighs_v_nr = adjs[id].size();
        update = dump * prev_pr_vals[id] / neighs_v_nr;

        for(v_id e = 0; e < neighs_v_nr; ++e) {
          v_id  neigh = adjs[id][e];
#pragma omp atomic
          pr_vals[neigh] += update;
        }

#pragma omp atomic
        pr_vals[id] += random_jump;

 #ifdef __XGRAPH_COUNT_OPERATIONS__
         //s.increase_atomic_stats(16 + neighs_v_nr*7, 9 + neighs_v_nr, 0, 0, neighs_v_nr);
         s.increase_atomic_stats(0, 0, neighs_v_nr);
 #endif
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_atomic_global_stats(&s_global, &s);
      s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_atomic_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
        //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
      }
#endif

#pragma omp barrier

#pragma omp single
      {
        sum = 0;
      }   

#pragma omp for reduction(+:sum)
      for (v_id i = 0; i < n; i++)
        sum += pr_vals[i];

#pragma omp single
      {
        sum = 1.0 / sum;
      }   

#pragma omp for 
      for (v_id i = 0; i < n; i++)
        pr_vals[i] *= sum;





#ifdef __BENCHMARK__
#pragma omp single
{
  LSB_Set_Rparam_string("phs", "conv");
  LSB_Res();
}
#endif

// #ifdef __XGRAPH_COUNT_OPERATIONS__
// papi_s.start_papi_daint_counters();
// #endif




#pragma omp single
      {
        err = 0;

/*        for (v_id i = 0; i < n; i++) {
          //if(fabs(pr_vals[i] - prev_pr_vals[i]) <= tolerance/(double)(n*n)) 
          if(fabs(pr_vals[i] - prev_pr_vals[i]) <= tolerance_v_conv) {
            converged[i] = true;
          }
        }
*/


        //to_remove.clear();

        for (v_id i = 0; i < n; i++) {
          //if(fabs(pr_vals[i] - prev_pr_vals[i]) <= tolerance/(double)(n*n)) 
          if(fabs(pr_vals[i] - prev_pr_vals[i]) <= tolerance_v_conv) {
            //to_remove(i);
            converged[i] = true;
          }
        }

      }   

#pragma omp for 
        for (v_id i = 0; i < n; i++) {
          new_adjs[i].clear();

          for(v_id w = 0; w < (v_id)adjs[i].size(); ++w) {
            v_id neighbor = adjs[i][w];
            if(!converged[neighbor]) {
              new_adjs[i].push_back(neighbor);
            }
          }

          adjs[i].clear();
          adjs[i] = new_adjs[i];
        }


/*#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);
#endif*/

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
/*#ifdef __XGRAPH_COUNT_OPERATIONS__
        papi_s_global.record_papi_in_lsb();
#endif*/

        LSB_Rec(run_id);
        //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
      }
#endif

#pragma omp barrier




/*#pragma omp for
        for (v_id i = 0; i < n; i++) {
          if(fabs(pr_vals[i] - prev_pr_vals[i]) <= tolerance/n) {
            _converged[i] = 1;
          }
        }
*/

#pragma omp for reduction(+:err)
      for (v_id i = 0; i < n; i++)
        err += fabs(pr_vals[i] - prev_pr_vals[i]);

      if (err < tolerance)
        break;

#pragma omp single
      { 
        ++iter;
      }
    }
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;

  delete [] adjs;

  // delete [] _converged;
}

//TODO: fix, it's broken
void PR::runPushLimitedAtomicsIfsWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prepare");
  LSB_Res();
#endif

  v_id n = G_csr->n_;

  double initial = 1.0 / n;
  double* pr_vals = new double[n];
  double* prev_pr_vals = new double[n];
  double random_jump = (1.0 - dump) / n;

#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    pr_vals[i] = initial;
    prev_pr_vals[i] = 0.0;
  }

  int iter = 1;

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  while (iter < maxiter) {
#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

#pragma omp parallel for
    for (v_id i = 0; i < n; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

#pragma omp parallel 
    {
      int thread_id = omp_get_thread_num();
      int threads_num = omp_get_num_threads();
#pragma omp for
      for (v_id id = 0; id < n; id++) {
        v_id* offsets = G_csr->offsets_;
        v_id* adj = G_csr->adj_;

        v_id id_offset = offsets[id];
        v_id neighs_v_nr = offsets[id+1] - offsets[id];
        v_id n_id = -1;
        double update = dump * prev_pr_vals[id] / neighs_v_nr;
        for(v_id e = 0; e < neighs_v_nr; ++e) {
          n_id = adj[id_offset + e];
          if(n_id % threads_num == thread_id) {
            pr_vals[n_id] += update;
          }
          else {
#pragma omp atomic
            pr_vals[n_id] += update;
          }
        }
        pr_vals[id] += random_jump;
      }
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
#pragma omp parallel for reduction(+:sum)
    for (v_id i = 0; i < n; i++)
      sum += pr_vals[i];

    sum = 1.0 / sum;
#pragma omp parallel for 
    for (v_id i = 0; i < n; i++)
      pr_vals[i] *= sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    double err = 0.0;
#pragma omp parallel for reduction(+:err)
    for (v_id i = 0; i < n; i++)
      err += fabs(pr_vals[i] - prev_pr_vals[i]);

    if (err < tolerance)
      break;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;
}


void PR::runPushLimitedAtomicsLocWithOpenMP(TradArrayLocalityAwareNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out) {
  v_id n = G_csr->n_;

  double initial = 1.0 / n;
  double* pr_vals = new double[n];
  double* prev_pr_vals = new double[n];
  double random_jump = (1.0 - dump) / n;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    pr_vals[i] = initial;
    prev_pr_vals[i] = 0.0;
  }

  int iter = 1;
  double sum = 0.0;
  double err = 0.0;

#pragma omp parallel 
  {



#pragma omp single
    {
      //v_id atomics = 0;
      cout << "N: " << n << ", M:" << G_csr->m_ << endl;

      v_id tot_atomics = 0;
      v_id tot_accesses = 0;
      v_id tot_m_l = 0;
      v_id tot_m_r = 0;

      for(int thread_id = 0; thread_id < omp_get_num_threads(); ++thread_id) {
        //v_id* my_adj_local = (G_csr->adj_local_)[thread_id];
        //v_id* my_adj_remote = (G_csr->adj_remote_)[thread_id];

        v_id* my_offsets_local = (G_csr->offsets_local_)[thread_id];
        v_id* my_offsets_remote = (G_csr->offsets_remote_)[thread_id];

        v_id my_n_local = (G_csr->n_local_)[thread_id];
        v_id atomics = 0;
        v_id accesses = 0;

        v_id m_l = (G_csr->m_local_)[thread_id];
        v_id m_r = (G_csr->m_remote_)[thread_id];

        tot_m_l += m_l;
        tot_m_r += m_r;

        cout << thread_id << "] N local: " << my_n_local << ", M local: " << m_l << ", M remote: " << m_r << ", twice m remote: " << 2*m_r << ", m_l + m_r: " << m_l + m_r << endl;

        for (v_id id = 0; id < my_n_local; ++id) {
          //v_id id_global = G_csr->getVertexGlobalID(id, thread_id);
          //v_id offset_remote = my_offsets_remote[id];
          v_id deg_local = my_offsets_local[id+1] - my_offsets_local[id];
          v_id deg_remote = my_offsets_remote[id+1] - my_offsets_remote[id];

          atomics += deg_remote;
          accesses += deg_local;
        }
        cout << thread_id << "] atomics: " << atomics << ", accesses: " << accesses << ", sum: " << atomics + accesses << endl;
        tot_atomics += atomics;
        tot_accesses += accesses;
      }

      cout << "tot atomics: " << tot_atomics << ", tot accesses: " << tot_accesses << ", tot M local: " << tot_m_l << ", tot M remote: " << tot_m_r << endl;

    }









    v_id id_global = 0;
    v_id offset_local = 0;
    v_id offset_remote = 0;
    v_id deg_local = 0;
    v_id deg_remote = 0;
    double update = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (iter < maxiter) {
#pragma omp single
      {
        for(int i = 0; i < 10; ++i) {
          cout << pr_vals[i] << " ";
        }
        cout << endl;
      }

#pragma omp for
      for (v_id i = 0; i < n; i++) {
        prev_pr_vals[i] = pr_vals[i];
        pr_vals[i] = 0.0;
      }

#ifdef __BENCHMARK__
#pragma omp single
      {
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
      int thread_id = PAPI_thread_id(); 
#else
      int thread_id = omp_get_thread_num();
#endif

      v_id* my_adj_local = (G_csr->adj_local_)[thread_id];
      v_id* my_adj_remote = (G_csr->adj_remote_)[thread_id];

      v_id* my_offsets_local = (G_csr->offsets_local_)[thread_id];
      v_id* my_offsets_remote = (G_csr->offsets_remote_)[thread_id];

      v_id my_n_local = (G_csr->n_local_)[thread_id];

      for (v_id id = 0; id < my_n_local; ++id) {
        id_global = G_csr->getVertexGlobalID(id, thread_id);
        offset_remote = my_offsets_remote[id];
        deg_local = my_offsets_local[id+1] - my_offsets_local[id];
        deg_remote = my_offsets_remote[id+1] - my_offsets_remote[id];
        update = dump * prev_pr_vals[id_global] / (deg_local + deg_remote);

        for(v_id e = 0; e < deg_remote; ++e) {
#pragma omp atomic
          pr_vals[my_adj_remote[offset_remote + e]] += update;
        }

#ifdef __XGRAPH_COUNT_OPERATIONS__
        s.increase_atomic_stats(0, 0, deg_remote);
#endif
      }

#pragma omp barrier

      for (v_id id = 0; id < my_n_local; ++id) {
        id_global = G_csr->getVertexGlobalID(id, thread_id);
        offset_local = my_offsets_local[id];
        deg_local = my_offsets_local[id+1] - my_offsets_local[id];
        deg_remote = my_offsets_remote[id+1] - my_offsets_remote[id];
        update = dump * prev_pr_vals[id_global] / (deg_local + deg_remote);

        for(v_id e = 0; e < deg_local; ++e) {
          pr_vals[my_adj_local[offset_local + e]] += update;
        }
        pr_vals[id_global] += random_jump;
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_atomic_global_stats(&s_global, &s);
      s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_atomic_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif
        LSB_Rec(run_id);
      }
#endif

#pragma omp barrier

#pragma omp single
      { 
        sum = 0.0;
      }

#pragma omp for reduction(+:sum)
      for (v_id i = 0; i < n; i++)
        sum += pr_vals[i];

#pragma omp single
      { 
        sum = 1.0 / sum;
      }

#pragma omp for 
      for (v_id i = 0; i < n; i++)
        pr_vals[i] *= sum;

#pragma omp single
      { 
        err = 0.0;
      }

#pragma omp for reduction(+:err)
      for (v_id i = 0; i < n; i++)
        err += fabs(pr_vals[i] - prev_pr_vals[i]);

      if (err < tolerance)
        break;

#pragma omp single
      { 
        ++iter;
      }
    }
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;
}

void PR::runPullWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out) {
  v_id n = G_csr->n_;

  double initial = 1.0 / n;
  double* pr_vals = new double[n];
  double* prev_pr_vals = new double[n];
  double random_jump = (1.0 - dump) / n;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel for
  for (v_id i = 0; i < n; i++) {
    pr_vals[i] = initial;
    prev_pr_vals[i] = 0.0;
  }

  int iter = 1;
  double sum = 0.0;
  double err = 0.0;

#pragma omp parallel 
  { 
    v_id* offsets = G_csr->offsets_;
    v_id* adj = G_csr->adj_;

    v_id neighs_v_nr = 0;
    v_id neighs_n_nr = 0;
    v_id neigh_id = 0;
    v_id id_offset = 0;

#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while (iter < maxiter) {
#pragma omp single
      {
        for(int i = 0; i < 10; ++i) {
          cout << pr_vals[i] << " ";
        }
        cout << endl;
      }

#pragma omp for
      for (v_id i = 0; i < n; i++) {
        prev_pr_vals[i] = pr_vals[i];
        pr_vals[i] = 0.0;
      }

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp for
      for (v_id id = 0; id < n; id++) {
        neighs_v_nr = offsets[id+1] - offsets[id];
        id_offset = offsets[id];

        for(v_id e = 0; e < neighs_v_nr; ++e) {
          neigh_id = adj[id_offset + e];
          neighs_n_nr = offsets[neigh_id+1] - offsets[neigh_id];

          pr_vals[id] += dump * prev_pr_vals[neigh_id] / neighs_n_nr;
        }
        pr_vals[id] += random_jump;
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_atomic_global_stats(&s_global, &s);
      s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_atomic_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
        //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
      }
#endif

#pragma omp barrier

#pragma omp single
      {
        sum = 0.0;
      }

#pragma omp for reduction(+:sum)
      for (v_id i = 0; i < n; i++)
        sum += pr_vals[i];

#pragma omp single
      {
        sum = 1.0 / sum;
      }

#pragma omp for 
      for (v_id i = 0; i < n; i++)
        pr_vals[i] *= sum;

#pragma omp single
      {
        err = 0.0;
      } 

#pragma omp for reduction(+:err)
      for (v_id i = 0; i < n; i++)
        err += fabs(pr_vals[i] - prev_pr_vals[i]);

      if (err < tolerance)
        break;

#pragma omp single
      {
        ++iter;
      }
    }
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;
}













void PR::setPRParams(double dump, int maxiter, double tolerance, v_id n, v_id n_local, v_id m) {
  dump_ = dump;
  maxiter_ = maxiter;
  tolerance_ = tolerance;
  n_ = n;
  n_local_ = n_local;
  m_ = m;

  initial_ = 1.0 / n;
  random_jump_ = (1.0 - dump_) / n;
}

void PR::initializePRVals(double* pr_vals, double* prev_pr_vals) {
  for (v_id i = 0; i < n_local_; i++) {
    pr_vals[i] = initial_;
    prev_pr_vals[i] = 0.0;
  }
}

/*v_id* PR::getNumberOfEdgesToFromEachProc(TradArrayDistGraphR* G) {
  v_id* edges_to_each_proc = new v_id[G->num_procs_]();

  for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
  v_id v_deg = G->getLocalVertexDegree(v_local_id);

  for(v_id e = 0; e < v_deg; ++e) {
  v_id v_neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e ];
  v_id v_neigh_owner = G->getVertexOwningRank(v_neigh_global_id);
  ++edges_to_each_proc[neigh_owner];
  }
  }

  return edges_to_each_proc;
  }*/


#ifdef __XGRAPH_DIST__

/////////////////////////////////////////////////// MPI MP (MPI+MPI) //////////////////////////////////////////////////

void PR::runPush_dist_MP_MPI_plus_MPI_Alltoallv(TradArrayDistGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prepare");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  double* pr_vals = new double[n_local_]();
  double* prev_pr_vals = new double[n_local_]();

  initializePRVals(pr_vals, prev_pr_vals);

  //TODO: what if some values exceed INT_MAX? May happen in high-degree vertices.
  int* send_counts = new int[G->num_procs_]();
  int* receive_counts = new int[G->num_procs_]();

  for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
    v_id v_deg = G->getLocalVertexDegree(v_local_id);
    for(v_id e = 0; e < v_deg; ++e) {
      v_id v_neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e ];
      v_id v_neigh_owner = G->getVertexOwningRank(v_neigh_global_id);
      if(v_neigh_owner != G->rank_) {
        ++send_counts[v_neigh_owner];
        ++receive_counts[v_neigh_owner];

        assert(send_counts[v_neigh_owner] < (INT_MAX - 1));
        assert(receive_counts[v_neigh_owner] < (INT_MAX - 1));
      }
    }
  }

  assert(send_counts[G->rank_] == 0);
  assert(receive_counts[G->rank_] == 0);

  int* send_displacements = new int[G->num_procs_]();
  int* receive_displacements = new int[G->num_procs_]();

  send_displacements[0] = 0;
  receive_displacements[0] = 0;

  for(int proc = 1; proc < G->num_procs_; ++proc) {
    assert((int64_t)send_displacements[proc - 1] + (int64_t)send_counts[proc - 1] < (INT_MAX - 1));
    assert((int64_t)receive_displacements[proc - 1] + (int64_t)receive_counts[proc - 1] < (INT_MAX - 1));
    send_displacements[proc] = send_displacements[proc - 1] + send_counts[proc - 1];
    receive_displacements[proc] = receive_displacements[proc - 1] + receive_counts[proc - 1];
  }

  v_id sum_of_outgoing_incoming_edges = 0;
  for(int i = 0; i < G->num_procs_; ++i) {
    sum_of_outgoing_incoming_edges += send_counts[i];
  }

  assert(sum_of_outgoing_incoming_edges == (send_displacements[G->num_procs_ - 1] + send_counts[G->num_procs_ - 1]));
  assert(sum_of_outgoing_incoming_edges == (receive_displacements[G->num_procs_ - 1] + receive_counts[G->num_procs_ - 1]));

  double* received_updates = new double[sum_of_outgoing_incoming_edges]();
  double* sent_updates = new double[sum_of_outgoing_incoming_edges]();

  int iter = 1;

  int* temp_send_offsets = new int[G->num_procs_]();
  v_id* processed_vertex = new v_id[G->num_procs_]();

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();


  while (iter < maxiter) {

    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comm_prep");
    LSB_Res();
#endif

    memset(&temp_send_offsets[0], 0, sizeof(int) * G->num_procs_);
    memset(&processed_vertex[0], 0, sizeof(v_id) * G->num_procs_);

    v_id neigh_owner = -1;
    v_id neigh_local_id = -1;
    v_id neigh_global_id = -1;

#ifdef __XGRAPH_DEBUGGING__
    for(v_id i = 0; i < sum_of_outgoing_incoming_edges; ++i) {
      sent_updates[i] = -1.0;
    }
#endif

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id v_deg = G->getLocalVertexDegree(v_local_id);
      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e];
        neigh_owner = G->getVertexOwningRank(neigh_global_id);
        if(neigh_owner != G->rank_) {
          assert( (send_displacements[neigh_owner] + temp_send_offsets[neigh_owner]) < sum_of_outgoing_incoming_edges);
          assert(temp_send_offsets[neigh_owner] < send_counts[neigh_owner]);
          sent_updates[send_displacements[neigh_owner] + temp_send_offsets[neigh_owner]] = update;
          ++temp_send_offsets[neigh_owner];
        }
      }
    }

#ifdef __XGRAPH_DEBUGGING__
    for(v_id i = 0; i < sum_of_outgoing_incoming_edges; ++i) {
      assert(sent_updates[i] >= 0.0);
    }
#endif

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comm");
    LSB_Res();
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    MPI_Alltoallv(&sent_updates[0], &send_counts[0], &send_displacements[0], MPI_DOUBLE, &received_updates[0], &receive_counts[0], &receive_displacements[0], MPI_DOUBLE, G->comm_);

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Alltoallv took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    //TODO: Communicate vertex IDs and thus reduce the complexity of the loops below.

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e];
        neigh_owner = G->getVertexOwningRank(neigh_global_id);
        neigh_local_id = G->getVertexLocalID(neigh_global_id);

        if(neigh_owner != G->rank_) {
          pr_vals[v_local_id] += received_updates[ receive_displacements[neigh_owner] + processed_vertex[neigh_owner] ];
          ++processed_vertex[neigh_owner];
        }
        else { //Perform local PR updates;
          neigh_local_id = G->getVertexLocalID(neigh_global_id);
          pr_vals[neigh_local_id] += update;
        }
      }
      pr_vals[v_local_id] += random_jump_;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P2 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    double err = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      err += fabs(pr_vals[i] - prev_pr_vals[i]);

    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    if (err < tolerance)
      break;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;

  delete [] received_updates;
  delete [] sent_updates;

  delete [] send_counts;
  delete [] send_displacements;
  delete [] receive_counts;
  delete [] receive_displacements;

  delete [] temp_send_offsets;
  delete [] processed_vertex;
}

/*void PR::runPullWith_MPI_MP_MPI_plus_MPI(TradArrayNoDegGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {

  }*/



/////////////////////////////////////////////////// MPI RMA //////////////////////////////////////////////////

// void PR::setMPIRMAConfig(MPI_Comm comm) {
//   comm_ = comm;
// }

void PR::runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prepare");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  // TODO: probably delete.
#ifdef __XGRAPH_DEBUGGING__
  v_id debug_n = 0;
  MPI_Allreduce(&n_local_, &debug_n, 1, MPI_INT64_T, MPI_SUM, G->comm_);
  assert(debug_n == n_);

  v_id* n_locals = new v_id[G->num_procs_]();
  for(v_id v = 0; v < n_; ++v) {
    ++n_locals[G->getVertexOwningRank(v)];
  }

  v_id n_sum = 0;
  for(int p = 0; p < G->num_procs_; ++p) {
    n_sum += n_locals[p];
  }

  assert(n_sum == n_);

  delete [] n_locals;
#endif

  double* prev_pr_vals = new double[n_local_]();
  double* pr_vals = NULL; 

  foMPI_Win pr_vals_win;
  foMPI_Win_allocate(n_local_*sizeof(double), sizeof(double), MPI_INFO_NULL, G->comm_, &pr_vals, &pr_vals_win);

  initializePRVals(pr_vals, prev_pr_vals);

  foMPI_Win_fence(0, pr_vals_win);

  int iter = 1;

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();
  //  if(G->rank_ == 0) cout << "Preparation took: " << t2 - t1 << endl;

  while (iter < maxiter) {
    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();
#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    //    if(G->rank_ == 0) cout << "RMA, itr " << iter << endl;
    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    foMPI_Win_fence(0, pr_vals_win);

    v_id neigh_rank = -1;
    v_id neigh_global_id = -1;
    v_id neigh_remote_win_offset = -1;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id* local_offsets = G->offsets_local_;
      v_id* local_adj = G->adj_local_;

      //v_id v_global_id = G->getVertexGlobalID(v_local_id, G->rank_);
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      v_id v_offset = local_offsets[v_local_id];

      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = local_adj[v_offset + e];
        neigh_remote_win_offset = G->getVertexLocalID(neigh_global_id);
        neigh_rank = G->getVertexOwningRank(neigh_global_id);
        foMPI_Accumulate(&update, 1, MPI_DOUBLE, neigh_rank, neigh_remote_win_offset, 1, MPI_DOUBLE, foMPI_SUM, pr_vals_win);
      }
      foMPI_Accumulate(&random_jump_, 1, MPI_DOUBLE, G->rank_, v_local_id, 1, MPI_DOUBLE, foMPI_SUM, pr_vals_win);
    }

    foMPI_Win_fence(0, pr_vals_win);

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Accumulates took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    //    if(G->rank_ == 0) cout << "Start normalize " << endl;
    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    /* TODO: Here the error is derived as a sum of abs. values. Below - a max.
       double err = 0.0;
       for (v_id i = 0; i < n_local_; i++)
       err += fabs(pr_vals[i] - prev_pr_vals[i]);

       if(G->rank_ == 0) cout << "b " << endl;
       double global_err = 0.0;
       MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

       if (err < tolerance)
       break;
     */

    double err = 0.0;
    double val = 0;
    for (v_id i = 0; i < n_local_; i++) {
      val = fabs(pr_vals[i] - prev_pr_vals[i]);
      if(val > err) {
        err = val;
      }
    }

    //    cout << G->rank_ << "] local max error: " << err << endl;

    //    if(G->rank_ == 0) cout << "Start converge " << endl;
    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_MAX, G->comm_);

    //    if(G->rank_ == 0) cout << "Max error: " << global_err << endl;

    if (global_err < tolerance)
      break;


#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  foMPI_Win_fence(0, pr_vals_win);

  double* pr_vals_copy = new double[n_]();
  memcpy(&pr_vals_copy[0], &pr_vals[0], sizeof(double) * n_local_);
  *pr_vals_out = pr_vals_copy;

  foMPI_Win_free(&pr_vals_win);

  delete [] prev_pr_vals;
}

void PR::runPull_dist_MPIRMA(TradArrayDistGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prep");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  // TODO: probably delete.
#ifdef __XGRAPH_DEBUGGING__
  v_id debug_n = 0;
  MPI_Allreduce(&n_local_, &debug_n, 1, MPI_INT64_T, MPI_SUM, G->comm_);
  assert(debug_n == n_);

  v_id* n_locals = new v_id[G->num_procs_]();
  for(v_id v = 0; v < n_; ++v) {
    ++n_locals[G->getVertexOwningRank(v)];
  }

  v_id n_sum = 0;
  for(int p = 0; p < G->num_procs_; ++p) {
    n_sum += n_locals[p];
  }

  assert(n_sum == n_);

  delete [] n_locals;
#endif

  double* pr_vals = new double[n_local_]();
  double* prev_pr_vals = NULL; 
  foMPI_Win pr_vals_win;
  foMPI_Win_allocate(n_local_*sizeof(double), sizeof(double), MPI_INFO_NULL, G->comm_, &prev_pr_vals, &pr_vals_win);

  int64_t* degs = NULL; 
  foMPI_Win degs_win;
  foMPI_Win_allocate(n_local_*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &degs, &degs_win);

  initializePRVals(pr_vals, prev_pr_vals);

  foMPI_Win_fence(0, pr_vals_win);

  for(int i = 0; i < n_local_; ++i) {
    degs[i] = (G->offsets_local_)[i+1] - (G->offsets_local_)[i];
  }

  foMPI_Win_fence(0, degs_win);

  int iter = 1;

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();
  //  if(G->rank_ == 0) cout << "Preparation took: " << t2 - t1 << endl;

  while (iter < maxiter) {
    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();
#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    //    if(G->rank_ == 0) cout << "RMA, itr " << iter << endl;
    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    foMPI_Win_fence(0, pr_vals_win);

    v_id neigh_rank = -1;
    v_id neigh_global_id = -1;
    v_id neigh_remote_win_offset = -1;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    //double update = 0.0; //dump * prev_pr_vals[v_local_id] / v_deg;
    double neigh_pr = 0.0;
    int64_t neigh_deg = 0;

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id* local_offsets = G->offsets_local_;
      v_id* local_adj = G->adj_local_;

      //v_id v_global_id = G->getVertexGlobalID(v_local_id, G->rank_);
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      v_id v_offset = local_offsets[v_local_id];

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = local_adj[v_offset + e];
        neigh_remote_win_offset = G->getVertexLocalID(neigh_global_id);
        neigh_rank = G->getVertexOwningRank(neigh_global_id);

        foMPI_Get(&neigh_deg, 1, MPI_INT64_T, neigh_rank, neigh_remote_win_offset, 1, MPI_INT64_T, degs_win);
        foMPI_Get(&neigh_pr, 1, MPI_DOUBLE, neigh_rank, neigh_remote_win_offset, 1, MPI_DOUBLE, pr_vals_win);
        foMPI_Win_flush(neigh_rank, pr_vals_win);

        pr_vals[v_local_id] += dump * neigh_pr / ((double)neigh_deg);
      }
      pr_vals[v_local_id] += random_jump_;
    }

    foMPI_Win_fence(0, pr_vals_win);

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Accumulates took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    //    if(G->rank_ == 0) cout << "Start normalize " << endl;
    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    /* TODO: Here the error is derived as a sum of abs. values. Below - a max.
       double err = 0.0;
       for (v_id i = 0; i < n_local_; i++)
       err += fabs(pr_vals[i] - prev_pr_vals[i]);

       if(G->rank_ == 0) cout << "b " << endl;
       double global_err = 0.0;
       MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

       if (err < tolerance)
       break;
     */

    double err = 0.0;
    double val = 0;
    for (v_id i = 0; i < n_local_; i++) {
      val = fabs(pr_vals[i] - prev_pr_vals[i]);
      if(val > err) {
        err = val;
      }
    }

    //    cout << G->rank_ << "] local max error: " << err << endl;

    //    if(G->rank_ == 0) cout << "Start converge " << endl;
    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_MAX, G->comm_);

    //  if(G->rank_ == 0) cout << "Max error: " << global_err << endl;

    if (global_err < tolerance)
      break;


#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  foMPI_Win_fence(0, pr_vals_win);

  double* pr_vals_copy = new double[n_]();
  memcpy(&pr_vals_copy[0], &pr_vals[0], sizeof(double) * n_local_);
  *pr_vals_out = pr_vals_copy;

  foMPI_Win_free(&pr_vals_win);
  foMPI_Win_free(&degs_win);

  delete [] pr_vals;
}


/*void PR::runPull_dist_MPIRMA(TradArrayNoDegGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {

  }*/




























void PR::runPush_dist_MP_MPI_plus_MPI_Alltoallv(TradArrayDistG500GraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prepare");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  double* pr_vals = new double[n_local_]();
  double* prev_pr_vals = new double[n_local_]();

  initializePRVals(pr_vals, prev_pr_vals);

  //TODO: what if some values exceed INT_MAX? May happen in high-degree vertices.
  int* send_counts = new int[G->num_procs_]();
  int* receive_counts = new int[G->num_procs_]();

  for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
    v_id v_deg = G->getLocalVertexDegree(v_local_id);
    for(v_id e = 0; e < v_deg; ++e) {
      v_id v_neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e ];
      v_id v_neigh_owner = G->getVertexOwningRank(v_neigh_global_id);
      if(v_neigh_owner != G->rank_) {
        ++send_counts[v_neigh_owner];
        ++receive_counts[v_neigh_owner];

        assert(send_counts[v_neigh_owner] < (INT_MAX - 1));
        assert(receive_counts[v_neigh_owner] < (INT_MAX - 1));
      }
    }
  }

  assert(send_counts[G->rank_] == 0);
  assert(receive_counts[G->rank_] == 0);

  int* send_displacements = new int[G->num_procs_]();
  int* receive_displacements = new int[G->num_procs_]();

  send_displacements[0] = 0;
  receive_displacements[0] = 0;

  for(int proc = 1; proc < G->num_procs_; ++proc) {
    assert((int64_t)send_displacements[proc - 1] + (int64_t)send_counts[proc - 1] < (INT_MAX - 1));
    assert((int64_t)receive_displacements[proc - 1] + (int64_t)receive_counts[proc - 1] < (INT_MAX - 1));
    send_displacements[proc] = send_displacements[proc - 1] + send_counts[proc - 1];
    receive_displacements[proc] = receive_displacements[proc - 1] + receive_counts[proc - 1];
  }

  v_id sum_of_outgoing_incoming_edges = 0;
  for(int i = 0; i < G->num_procs_; ++i) {
    sum_of_outgoing_incoming_edges += send_counts[i];
  }

  assert(sum_of_outgoing_incoming_edges == (send_displacements[G->num_procs_ - 1] + send_counts[G->num_procs_ - 1]));
  assert(sum_of_outgoing_incoming_edges == (receive_displacements[G->num_procs_ - 1] + receive_counts[G->num_procs_ - 1]));

  double* received_updates = new double[sum_of_outgoing_incoming_edges]();
  double* sent_updates = new double[sum_of_outgoing_incoming_edges]();

  int iter = 1;

  int* temp_send_offsets = new int[G->num_procs_]();
  v_id* processed_vertex = new v_id[G->num_procs_]();

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();


  while (iter < maxiter) {

    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comm_prep");
    LSB_Res();
#endif

    memset(&temp_send_offsets[0], 0, sizeof(int) * G->num_procs_);
    memset(&processed_vertex[0], 0, sizeof(v_id) * G->num_procs_);

    v_id neigh_owner = -1;
    v_id neigh_local_id = -1;
    v_id neigh_global_id = -1;

#ifdef __XGRAPH_DEBUGGING__
    for(v_id i = 0; i < sum_of_outgoing_incoming_edges; ++i) {
      sent_updates[i] = -1.0;
    }
#endif

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id v_deg = G->getLocalVertexDegree(v_local_id);
      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e];
        neigh_owner = G->getVertexOwningRank(neigh_global_id);
        if(neigh_owner != G->rank_) {
          assert( (send_displacements[neigh_owner] + temp_send_offsets[neigh_owner]) < sum_of_outgoing_incoming_edges);
          assert(temp_send_offsets[neigh_owner] < send_counts[neigh_owner]);
          sent_updates[send_displacements[neigh_owner] + temp_send_offsets[neigh_owner]] = update;
          ++temp_send_offsets[neigh_owner];
        }
      }
    }

#ifdef __XGRAPH_DEBUGGING__
    for(v_id i = 0; i < sum_of_outgoing_incoming_edges; ++i) {
      assert(sent_updates[i] >= 0.0);
    }
#endif

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comm");
    LSB_Res();
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    MPI_Alltoallv(&sent_updates[0], &send_counts[0], &send_displacements[0], MPI_DOUBLE, &received_updates[0], &receive_counts[0], &receive_displacements[0], MPI_DOUBLE, G->comm_);

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Alltoallv took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    //TODO: Communicate vertex IDs and thus reduce the complexity of the loops below.

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = (G->adj_local_)[ (G->offsets_local_)[v_local_id] + e];
        neigh_owner = G->getVertexOwningRank(neigh_global_id);
        neigh_local_id = G->getVertexLocalID(neigh_global_id);

        if(neigh_owner != G->rank_) {
          pr_vals[v_local_id] += received_updates[ receive_displacements[neigh_owner] + processed_vertex[neigh_owner] ];
          ++processed_vertex[neigh_owner];
        }
        else { //Perform local PR updates;
          neigh_local_id = G->getVertexLocalID(neigh_global_id);
          pr_vals[neigh_local_id] += update;
        }
      }
      pr_vals[v_local_id] += random_jump_;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P2 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    double err = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      err += fabs(pr_vals[i] - prev_pr_vals[i]);

    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    if (err < tolerance)
      break;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  *pr_vals_out = pr_vals;

  delete [] prev_pr_vals;

  delete [] received_updates;
  delete [] sent_updates;

  delete [] send_counts;
  delete [] send_displacements;
  delete [] receive_counts;
  delete [] receive_displacements;

  delete [] temp_send_offsets;
  delete [] processed_vertex;
}





void PR::runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistG500GraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prepare");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  // TODO: probably delete.
#ifdef __XGRAPH_DEBUGGING__
  v_id debug_n = 0;
  MPI_Allreduce(&n_local_, &debug_n, 1, MPI_INT64_T, MPI_SUM, G->comm_);
  assert(debug_n == n_);

  v_id* n_locals = new v_id[G->num_procs_]();
  for(v_id v = 0; v < n_; ++v) {
    ++n_locals[G->getVertexOwningRank(v)];
  }

  v_id n_sum = 0;
  for(int p = 0; p < G->num_procs_; ++p) {
    n_sum += n_locals[p];
  }

  assert(n_sum == n_);

  delete [] n_locals;
#endif

  double* prev_pr_vals = new double[n_local_]();
  double* pr_vals = NULL; 

  foMPI_Win pr_vals_win;
  foMPI_Win_allocate(n_local_*sizeof(double), sizeof(double), MPI_INFO_NULL, G->comm_, &pr_vals, &pr_vals_win);

  initializePRVals(pr_vals, prev_pr_vals);

  foMPI_Win_fence(0, pr_vals_win);

  int iter = 1;

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();
  //  if(G->rank_ == 0) cout << "Preparation took: " << t2 - t1 << endl;

  while (iter < maxiter) {
    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();
#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    //    if(G->rank_ == 0) cout << "RMA, itr " << iter << endl;
    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    foMPI_Win_fence(0, pr_vals_win);

    v_id neigh_rank = -1;
    v_id neigh_global_id = -1;
    v_id neigh_remote_win_offset = -1;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id* local_offsets = G->offsets_local_;
      v_id* local_adj = G->adj_local_;

      //v_id v_global_id = G->getVertexGlobalID(v_local_id, G->rank_);
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      v_id v_offset = local_offsets[v_local_id];

      double update = dump * prev_pr_vals[v_local_id] / v_deg;

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = local_adj[v_offset + e];
        neigh_remote_win_offset = G->getVertexLocalID(neigh_global_id);
        neigh_rank = G->getVertexOwningRank(neigh_global_id);
        foMPI_Accumulate(&update, 1, MPI_DOUBLE, neigh_rank, neigh_remote_win_offset, 1, MPI_DOUBLE, foMPI_SUM, pr_vals_win);
      }
      foMPI_Accumulate(&random_jump_, 1, MPI_DOUBLE, G->rank_, v_local_id, 1, MPI_DOUBLE, foMPI_SUM, pr_vals_win);
    }

    foMPI_Win_fence(0, pr_vals_win);

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Accumulates took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    //    if(G->rank_ == 0) cout << "Start normalize " << endl;
    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    /* TODO: Here the error is derived as a sum of abs. values. Below - a max.
       double err = 0.0;
       for (v_id i = 0; i < n_local_; i++)
       err += fabs(pr_vals[i] - prev_pr_vals[i]);

       if(G->rank_ == 0) cout << "b " << endl;
       double global_err = 0.0;
       MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

       if (err < tolerance)
       break;
     */

    double err = 0.0;
    double val = 0;
    for (v_id i = 0; i < n_local_; i++) {
      val = fabs(pr_vals[i] - prev_pr_vals[i]);
      if(val > err) {
        err = val;
      }
    }

    //    cout << G->rank_ << "] local max error: " << err << endl;

    //    if(G->rank_ == 0) cout << "Start converge " << endl;
    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_MAX, G->comm_);

    //    if(G->rank_ == 0) cout << "Max error: " << global_err << endl;

    if (global_err < tolerance)
      break;


#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  foMPI_Win_fence(0, pr_vals_win);

  double* pr_vals_copy = new double[n_]();
  memcpy(&pr_vals_copy[0], &pr_vals[0], sizeof(double) * n_local_);
  *pr_vals_out = pr_vals_copy;

  foMPI_Win_free(&pr_vals_win);

  delete [] prev_pr_vals;
}

void PR::runPull_dist_MPIRMA(TradArrayDistG500GraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out) {
#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prep");
  LSB_Res();
#endif

  double t1, t2;
  t1 = MPI_Wtime();

  setPRParams(dump, maxiter, tolerance, G->n_, G->n_local_, G->m_);

  // TODO: probably delete.
#ifdef __XGRAPH_DEBUGGING__
  v_id debug_n = 0;
  MPI_Allreduce(&n_local_, &debug_n, 1, MPI_INT64_T, MPI_SUM, G->comm_);
  assert(debug_n == n_);

  v_id* n_locals = new v_id[G->num_procs_]();
  for(v_id v = 0; v < n_; ++v) {
    ++n_locals[G->getVertexOwningRank(v)];
  }

  v_id n_sum = 0;
  for(int p = 0; p < G->num_procs_; ++p) {
    n_sum += n_locals[p];
  }

  assert(n_sum == n_);

  delete [] n_locals;
#endif

  double* pr_vals = new double[n_local_]();
  double* prev_pr_vals = NULL; 
  foMPI_Win pr_vals_win;
  foMPI_Win_allocate(n_local_*sizeof(double), sizeof(double), MPI_INFO_NULL, G->comm_, &prev_pr_vals, &pr_vals_win);

  int64_t* degs = NULL; 
  foMPI_Win degs_win;
  foMPI_Win_allocate(n_local_*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &degs, &degs_win);

  initializePRVals(pr_vals, prev_pr_vals);

  foMPI_Win_fence(0, pr_vals_win);

  for(int i = 0; i < n_local_; ++i) {
    degs[i] = (G->offsets_local_)[i+1] - (G->offsets_local_)[i];
  }

  foMPI_Win_fence(0, degs_win);

  int iter = 1;

#ifdef __BENCHMARK__
  LSB_Rec(run_id);
#endif

  t2 = MPI_Wtime();
  //  if(G->rank_ == 0) cout << "Preparation took: " << t2 - t1 << endl;

  while (iter < maxiter) {
    if(G->rank_ == 0) {
      for(int i = 0; i < 10; ++i) {
        cout << pr_vals[i] << " ";
      }
      cout << endl;
    }

    t1 = MPI_Wtime();
#ifdef __BENCHMARK__
    LSB_Set_Rparam_int("itr", iter);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "swap");
    LSB_Res();
#endif

    //    if(G->rank_ == 0) cout << "RMA, itr " << iter << endl;
    for (v_id i = 0; i < n_local_; i++) {
      prev_pr_vals[i] = pr_vals[i];
      pr_vals[i] = 0.0;
    }

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
#endif

    foMPI_Win_fence(0, pr_vals_win);

    v_id neigh_rank = -1;
    v_id neigh_global_id = -1;
    v_id neigh_remote_win_offset = -1;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "P1 took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

    //double update = 0.0; //dump * prev_pr_vals[v_local_id] / v_deg;
    double neigh_pr = 0.0;
    int64_t neigh_deg = 0;

    for (v_id v_local_id = 0; v_local_id < n_local_; ++v_local_id) {
      v_id* local_offsets = G->offsets_local_;
      v_id* local_adj = G->adj_local_;

      //v_id v_global_id = G->getVertexGlobalID(v_local_id, G->rank_);
      v_id v_deg = G->getLocalVertexDegree(v_local_id);

      v_id v_offset = local_offsets[v_local_id];

      for(v_id e = 0; e < v_deg; ++e) {
        neigh_global_id = local_adj[v_offset + e];
        neigh_remote_win_offset = G->getVertexLocalID(neigh_global_id);
        neigh_rank = G->getVertexOwningRank(neigh_global_id);

        foMPI_Get(&neigh_deg, 1, MPI_INT64_T, neigh_rank, neigh_remote_win_offset, 1, MPI_INT64_T, degs_win);
        foMPI_Get(&neigh_pr, 1, MPI_DOUBLE, neigh_rank, neigh_remote_win_offset, 1, MPI_DOUBLE, pr_vals_win);
        foMPI_Win_flush(neigh_rank, pr_vals_win);

        pr_vals[v_local_id] += dump * neigh_pr / ((double)neigh_deg);
      }
      pr_vals[v_local_id] += random_jump_;
    }

    foMPI_Win_fence(0, pr_vals_win);

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Accumulates took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "norm");
    LSB_Res();
#endif

    double sum = 0.0;
    for (v_id i = 0; i < n_local_; i++)
      sum += pr_vals[i];

    //    if(G->rank_ == 0) cout << "Start normalize " << endl;
    double global_sum = 0.0;
    MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

    global_sum = 1.0 / global_sum;
    for (v_id i = 0; i < n_local_; i++)
      pr_vals[i] *= global_sum;

#ifdef __BENCHMARK__
    LSB_Rec(run_id);
#endif

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Norm took: " << t2 - t1 << endl;

    t1 = MPI_Wtime();

#ifdef __BENCHMARK__
    LSB_Set_Rparam_string("phs", "conv");
    LSB_Res();
#endif

    /* TODO: Here the error is derived as a sum of abs. values. Below - a max.
       double err = 0.0;
       for (v_id i = 0; i < n_local_; i++)
       err += fabs(pr_vals[i] - prev_pr_vals[i]);

       if(G->rank_ == 0) cout << "b " << endl;
       double global_err = 0.0;
       MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_SUM, G->comm_);

       if (err < tolerance)
       break;
     */

    double err = 0.0;
    double val = 0;
    for (v_id i = 0; i < n_local_; i++) {
      val = fabs(pr_vals[i] - prev_pr_vals[i]);
      if(val > err) {
        err = val;
      }
    }

    //    cout << G->rank_ << "] local max error: " << err << endl;

    //    if(G->rank_ == 0) cout << "Start converge " << endl;
    double global_err = 0.0;
    MPI_Allreduce(&err, &global_err, 1, MPI_DOUBLE, MPI_MAX, G->comm_);

    //  if(G->rank_ == 0) cout << "Max error: " << global_err << endl;

    if (global_err < tolerance)
      break;


#ifdef __BENCHMARK__
    LSB_Rec(run_id);
    //if(iter % __XGRAPH_LSB_FLUSH_FREQUENCY == 0) {LSB_Flush();}
#endif

    ++iter;

    t2 = MPI_Wtime();
    //    if(G->rank_ == 0) cout << "Conv took: " << t2 - t1 << endl;
  }

  foMPI_Win_fence(0, pr_vals_win);

  double* pr_vals_copy = new double[n_]();
  memcpy(&pr_vals_copy[0], &pr_vals[0], sizeof(double) * n_local_);
  *pr_vals_out = pr_vals_copy;

  foMPI_Win_free(&pr_vals_win);
  foMPI_Win_free(&degs_win);

  delete [] pr_vals;
}










#endif
