#if !defined(__PR_H__)
#define      __PR_H__

#ifdef __XGRAPH_DIST__
#include "fompi.h"
#include "representations/tradArrayDistGraphR.h"
#include "representations/tradArrayDistG500GraphR.h"
#include "representations/external_generators/g500_rmat/common.h"
#endif

#include "globals.h"
#include "utils/graphIO.h"

#ifdef __XGRAPH_COUNT_OPERATIONS__
#include "utils/stats.h"
#endif

#include "representations/tradArrayNoDegGraphR.h"
#include "representations/tradArrayLocalityAwareNoDegGraphR.h"

using namespace std;

class PR {
 public:
  void setOpenmpConfig(int dynamic, int threads);
  void runPushWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pagerank_out);
  void runPushConvWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double tolerance_v_conv, double** pagerank_out);
  void runPushLimitedAtomicsIfsWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out);
  void runPushLimitedAtomicsLocWithOpenMP(TradArrayLocalityAwareNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pr_vals_out);
  void runPullWithOpenMP(TradArrayNoDegGraphR* G_csr, double dump, int maxiter, double tolerance, double** pagerank_out);

#ifdef __XGRAPH_DIST__
  void runPull_dist_MPIRMA(TradArrayDistGraphR* G_csr, double dump, int maxiter, double tolerance, double** pagerank_out);
  void runPush_dist_MP_MPI_plus_MPI_Alltoallv(TradArrayDistGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out);
  void runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistGraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out);

  void runPull_dist_MPIRMA(TradArrayDistG500GraphR* G_csr, double dump, int maxiter, double tolerance, double** pagerank_out);
  void runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistG500GraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out);
  void runPush_dist_MP_MPI_plus_MPI_Alltoallv(TradArrayDistG500GraphR* G, double dump, int maxiter, double tolerance, double** pr_vals_out);
#endif

  void initializePRVals(double* pr_vals, double* prev_pr_vals);
  void setPRParams(double dump, int maxiter, double tolerance, v_id n, v_id n_local, v_id m);

  void setRunID(int run_id) {this->run_id = run_id;}

  int run_id;

#ifdef __XGRAPH_COUNT_OPERATIONS__
  int papi_set;
#endif

  double dump_;
  int maxiter_;
  double tolerance_;
          
  double initial_;
  double random_jump_;
  v_id n_;
  v_id n_local_;
  v_id m_;
};

#endif
