#include "TC.h"
#include <array>

void TC::setOpenmpConfig(int dynamic, int threads) {
  omp_set_dynamic(dynamic);
  omp_set_num_threads(threads);
}

void TC::runPushWithOpenMP(TradArrayNoDegGraphR* G_csr, int64_t** tc_vals_out) {
  v_id n = G_csr->n_;
  int64_t* tc_vals = new int64_t[n]();

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel 
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

#ifdef __BENCHMARK__
#pragma omp single
    { 
      LSB_Set_Rparam_string("phs", "comp");
      LSB_Res();
    }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.start_papi_daint_counters();
#endif

#pragma omp for
    for (v_id id = 0; id < n; ++id) {

      v_id* offsets = G_csr->offsets_;
      v_id* adj = G_csr->adj_;

      v_id id_offset = offsets[id];
      v_id neighs_v_nr = offsets[id + 1] - offsets[id];

      for (v_id i = 0; i < neighs_v_nr; ++i) {

        v_id id_offset_n = offsets[adj[id_offset + i]];
        v_id neighs_v_nr_n = offsets[adj[id_offset + i] + 1] - offsets[adj[id_offset + i]];

        for (v_id it = 0; it < neighs_v_nr; ++it) {
          for (v_id e = 0; e < neighs_v_nr_n; ++e) {
            if (adj[id_offset + it] == adj[id_offset_n + e]){
              __sync_fetch_and_add(&tc_vals[adj[id_offset + i]], 1);

#ifdef __XGRAPH_COUNT_OPERATIONS__
              ++s.fad_;
#endif
            }
          }
        }
      }
    }

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.stop_papi_daint_counters();
    PapiStats::increase_global_stats(&papi_s_global, &papi_s);

    MemoryStats::increase_atomic_global_stats(&s_global, &s);
    s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
    { 
#ifdef __XGRAPH_COUNT_OPERATIONS__
      s_global.record_atomic_stats_in_lsb();
      papi_s_global.record_papi_in_lsb();
#endif

      LSB_Rec(run_id);
    }
#endif
  }

  // it is not necessary if we are using sorted tables
#pragma omp parallel for
  for (unsigned i = 0; i < n; ++i) {
    tc_vals[i] = tc_vals[i]/2;
  }

  *tc_vals_out = tc_vals;

  for(int i = 0; i < 25; ++i) {
    cout << (*tc_vals_out)[i] << " ";
  }
  cout << endl;

}

void TC::runPullWithOpenMP(TradArrayNoDegGraphR* G_csr, int64_t** tc_vals_out) {
  v_id n = G_csr->n_;
  int64_t* tc_vals = new int64_t[n]();

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

#pragma omp parallel 
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

#ifdef __BENCHMARK__
#pragma omp single
    { 
      LSB_Set_Rparam_string("phs", "comp");
      LSB_Res();
    }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.start_papi_daint_counters();
#endif

#pragma omp for
    for (v_id id = 0; id < n; ++id) {

      v_id* offsets = G_csr->offsets_;
      v_id* adj = G_csr->adj_;

      v_id id_offset = offsets[id];
      v_id neighs_v_nr = offsets[id + 1] - offsets[id];

      for (v_id i = 0; i < neighs_v_nr; ++i) {

        v_id id_offset_n = offsets[adj[id_offset + i]];
        v_id neighs_v_nr_n = offsets[adj[id_offset + i] + 1] - offsets[adj[id_offset + i]];

        for (v_id it = 0; it < neighs_v_nr; ++it) {
          for (v_id e = 0; e < neighs_v_nr_n; ++e) {
            if (adj[id_offset + it] == adj[id_offset_n + e]){
              ++tc_vals[id];
            }
          }
        }
      }
    }

#ifdef __XGRAPH_COUNT_OPERATIONS__
    papi_s.stop_papi_daint_counters();
    PapiStats::increase_global_stats(&papi_s_global, &papi_s);

    MemoryStats::increase_atomic_global_stats(&s_global, &s);
    s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
    {
#ifdef __XGRAPH_COUNT_OPERATIONS__
      s_global.record_atomic_stats_in_lsb();
      papi_s_global.record_papi_in_lsb();
#endif

      LSB_Rec(run_id);
    }
#endif
  }

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n; ++i) {
    tc_vals[i] = tc_vals[i]/2;
  }

  *tc_vals_out = tc_vals;
}


#ifdef __XGRAPH_DIST__


void TC::runPush_dist_MP_MPI(TradArrayDistGraphR* G, int64_t** tc_vals_out) {

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
#endif

#ifdef __BENCHMARK__
  LSB_Set_Rparam_int("p", G->rank_);
  LSB_Set_Rparam_int("itr", 0);
  LSB_Set_Rparam_string("phs", "prep");
  LSB_Res();
#endif

  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* tc_vals = new int64_t[n_local](); 

  MPI_Barrier(G->comm_);

  /*

#ifdef __BENCHMARK__
LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
papi_s.stop_papi_daint_counters();
papi_s.record_papi_in_lsb();

s.record_remote_stats_in_lsb();
s.zero_all_stats();
#endif

LSB_Set_Rparam_string("phs", "comp");
LSB_Res();

#ifdef __XGRAPH_COUNT_OPERATIONS__
papi_s.start_papi_daint_counters();
#endif
#endif

v_id* offsets_local = G->offsets_local_;
v_id* adj_local = G->adj_local_;

v_id v_l_offset = -1;
v_id v_l_deg = -1;
v_id neigh_id_g = -1;
v_id neigh_offset = -1;
v_id neigh_offset_next = -1;
v_id neigh_deg = -1;

v_id neigh_id_l = -1;
int neigh_owner = -1;

for (v_id v_l = 0; v_l < n_local; ++v_l) {
v_l_offset = offsets_local[v_l];
v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

for (v_id i = 0; i < v_l_deg; ++i) {
neigh_id_g = adj_local[v_l_offset + i];
neigh_id_l = G->getVertexLocalID(neigh_id_g);
neigh_owner = G->getVertexOwningRank(neigh_id_g);

if(neigh_owner == G->rank_) {
neigh_offset = offsets_local[neigh_id_l];
neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

for (v_id it = 0; it < v_l_deg; ++it) {
for (v_id e = 0; e < neigh_deg; ++e) {
if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
  //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
  int64_t add = 1;
  int64_t res = 0;
  foMPI_Get_accumulate(&add, 1, MPI_INT64_T, &res, 1, MPI_INT64_T, neigh_owner, neigh_id_l, 1, MPI_INT64_T, foMPI_SUM, tc_vals_win);

#ifdef __XGRAPH_COUNT_OPERATIONS__
++s.local_fompi_gac_sum_;
#endif
}
}
}
}
else {
int64_t data[2];
foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
foMPI_Win_flush(neigh_owner, offsets_local_win);

neigh_offset = data[0];
neigh_offset_next = data[1];
neigh_deg = neigh_offset_next - neigh_offset;

  //TODO: does it consume too much memory?
  v_id* neighbors = new v_id[neigh_deg]();

  foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
  foMPI_Win_flush(neigh_owner, adj_local_win);

  v_id adj_local_1 = -1;
  v_id adj_local_2 = -1;
  for (v_id it = 0; it < v_l_deg; ++it) {
    adj_local_1 = adj_local[v_l_offset + it];

    for (v_id e = 0; e < neigh_deg; ++e) {
      adj_local_2 = neighbors[e];

      if (adj_local_1 == adj_local_2){
        //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
        int64_t add = 1;
        int64_t res = 0;
        v_id local_id = G->getLocalVertexDegree(adj_local_2);
        foMPI_Get_accumulate(&add, 1, MPI_INT64_T, &res, 1, MPI_INT64_T, neigh_owner, neigh_id_l, 1, MPI_INT64_T, foMPI_SUM, tc_vals_win);

#ifdef __XGRAPH_COUNT_OPERATIONS__
        ++s.remote_fompi_gac_sum_;
#endif
      }
    }
  }
}
}
}

foMPI_Win_fence(0, tc_vals_win);

#ifdef __BENCHMARK__
LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
papi_s.stop_papi_daint_counters();
papi_s.record_papi_in_lsb();

s.record_remote_stats_in_lsb();
s.zero_all_stats();
#endif
#endif

// it is not necessary if we are using sorted tables
for (unsigned i = 0; i < n; ++i) {
  tc_vals[i] = tc_vals[i]/2;
}
*/

*tc_vals_out = tc_vals;

}





















void TC::runPull_dist_MPIRMA(TradArrayDistG500GraphR* G, int64_t** tc_vals_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
  s.zero_all_stats();
#endif

  //cout << "...a" << endl;

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Set_Rparam_int("p", G->rank_);
    LSB_Set_Rparam_int("itr", 0);
    LSB_Set_Rparam_string("phs", "prep");
    LSB_Res();
  }
#endif

  //cout << "...b" << endl;
  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* local_tc_values = new int64_t[n_local](); 
  //int64_t* tc_vals = NULL; 

  //foMPI_Win tc_vals_win;
  //foMPI_Win_allocate(n_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &tc_vals, &tc_vals_win);

  //cout << "...c" << endl;
  foMPI_Win offsets_local_win;
  foMPI_Win adj_local_win;
  foMPI_Win_create(G->offsets_local_, (n_local+1)*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &offsets_local_win);
  //cout << "...d" << endl;
  foMPI_Win_create(G->adj_local_, m_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &adj_local_win);
  //cout << "...e" << endl;

  //foMPI_Win_fence(0, tc_vals_win);
  foMPI_Win_fence(0, adj_local_win);

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);

    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
  }
#endif

  v_id* offsets_local = G->offsets_local_;
  v_id* adj_local = G->adj_local_;

  v_id v_l_offset = -1;
  v_id v_l_deg = -1;
  v_id neigh_id_g = -1;
  v_id neigh_offset = -1;
  v_id neigh_offset_next = -1;
  v_id neigh_deg = -1;

  v_id neigh_id_l = -1;
  int neigh_owner = -1;

  //cout << "...f" << endl;
  for (v_id v_l = 0; v_l < n_local; ++v_l) {
    v_l_offset = offsets_local[v_l];
    v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

    for (v_id i = 0; i < v_l_deg; ++i) {
      neigh_id_g = adj_local[v_l_offset + i];
      neigh_id_l = G->getVertexLocalID(neigh_id_g);
      neigh_owner = G->getVertexOwningRank(neigh_id_g);

      if(neigh_owner == G->rank_) {
        neigh_offset = offsets_local[neigh_id_l];
        neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

        for (v_id it = 0; it < v_l_deg; ++it) {
          for (v_id e = 0; e < neigh_deg; ++e) {
            if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);

              ++(local_tc_values[v_l]);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }
      }
      else {
        int64_t data[2];
        foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
        foMPI_Win_flush(neigh_owner, offsets_local_win);

        neigh_offset = data[0];
        neigh_offset_next = data[1];
        neigh_deg = neigh_offset_next - neigh_offset;

        //TODO: does it consume too much memory?
        v_id* neighbors = new v_id[neigh_deg]();

        foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
        foMPI_Win_flush(neigh_owner, adj_local_win);

        v_id adj_local_1 = -1;
        v_id adj_local_2 = -1;
        for (v_id it = 0; it < v_l_deg; ++it) {
          adj_local_1 = adj_local[v_l_offset + it];

          for (v_id e = 0; e < neigh_deg; ++e) {
            adj_local_2 = neighbors[e];

            if (adj_local_1 == adj_local_2){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              ++(local_tc_values[v_l]);
#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }

        delete [] neighbors;
      }
    }
  }

  //foMPI_Win_fence(0, tc_vals_win);
  foMPI_Win_fence(0, adj_local_win);

  if(G->rank_ == 0) cout << "Finished to compute TCs" << endl;

#ifdef __BENCHMARK__

  if(G->rank_ == 0) {
    LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
    s.record_remote_stats_in_lsb();
    s.zero_all_stats();
#endif
  }
#endif

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n_local; ++i) {
    local_tc_values[i] = local_tc_values[i]/2;
  }

  int64_t* tc_vals_copy = new int64_t[n_local]();
  memcpy(&tc_vals_copy[0], &local_tc_values[0], sizeof(int64_t) * n_local);

  *tc_vals_out = tc_vals_copy;

  delete [] local_tc_values;

  foMPI_Win_free(&adj_local_win);
  foMPI_Win_free(&offsets_local_win);

  if(G->rank_ == 0) cout << "Finishing..." << endl;
  if(G->rank_ == 0) {
    for(int i = 0; i < 25; ++i) {
      cout << (*tc_vals_out)[i] << " ";
    }
    cout << endl;
  }
}



void TC::runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistG500GraphR* G, int64_t** tc_vals_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
  s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Set_Rparam_int("p", G->rank_);
    LSB_Set_Rparam_int("itr", 0);
    LSB_Set_Rparam_string("phs", "prep");
    LSB_Res();
  }
#endif

  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* tc_vals = NULL; 

  foMPI_Win tc_vals_win;
  foMPI_Win_allocate(n_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &tc_vals, &tc_vals_win);

  foMPI_Win offsets_local_win;
  foMPI_Win adj_local_win;
  foMPI_Win_create(G->offsets_local_, (n_local+1)*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &offsets_local_win);
  foMPI_Win_create(G->adj_local_, m_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &adj_local_win);

  int64_t* local_tc_values = new int64_t[n_local]();

  foMPI_Win_fence(0, tc_vals_win);

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);
    LSB_Set_Rparam_string("phs", "comp");

    LSB_Res();
  }
#endif

  v_id* offsets_local = G->offsets_local_;
  v_id* adj_local = G->adj_local_;

  v_id v_l_offset = -1;
  v_id v_l_deg = -1;
  v_id neigh_id_g = -1;
  v_id neigh_offset = -1;
  v_id neigh_offset_next = -1;
  v_id neigh_deg = -1;

  v_id neigh_id_l = -1;
  int neigh_owner = -1;

  MPI_Barrier(G->comm_);

  if(G->rank_ == 0) cout << "Start to compute TCs" << endl;

  for (v_id v_l = 0; v_l < n_local; ++v_l) {
    v_l_offset = offsets_local[v_l];
    v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

    for (v_id i = 0; i < v_l_deg; ++i) {
      neigh_id_g = adj_local[v_l_offset + i];
      neigh_id_l = G->getVertexLocalID(neigh_id_g);
      neigh_owner = G->getVertexOwningRank(neigh_id_g);

      if(neigh_owner == G->rank_) {
        neigh_offset = offsets_local[neigh_id_l];
        neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

        for (v_id it = 0; it < v_l_deg; ++it) {
          for (v_id e = 0; e < neigh_deg; ++e) {
            if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              ++(local_tc_values[neigh_id_l]);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }
      }
      else {
        int64_t data[2];
        foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
        foMPI_Win_flush(neigh_owner, offsets_local_win);

        neigh_offset = data[0];
        neigh_offset_next = data[1];
        neigh_deg = neigh_offset_next - neigh_offset;

        //TODO: does it consume too much memory?
        v_id* neighbors = new v_id[neigh_deg]();

        foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
        foMPI_Win_flush(neigh_owner, adj_local_win);

        v_id adj_local_1 = -1;
        v_id adj_local_2 = -1;
        for (v_id it = 0; it < v_l_deg; ++it) {
          adj_local_1 = adj_local[v_l_offset + it];

          for (v_id e = 0; e < neigh_deg; ++e) {
            adj_local_2 = neighbors[e];

            if (adj_local_1 == adj_local_2){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              int64_t add = 1;
              int64_t res = 0;
              v_id local_id = G->getLocalVertexDegree(adj_local_2);
              foMPI_Get_accumulate(&add, 1, MPI_INT64_T, &res, 1, MPI_INT64_T, neigh_owner, neigh_id_l, 1, MPI_INT64_T, foMPI_SUM, tc_vals_win);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.remote_fompi_gac_sum_;
  }
#endif
            }
          }
        }
        delete [] neighbors;
      }
    }
  }

  foMPI_Win_fence(0, tc_vals_win);

  if(G->rank_ == 0) cout << "Finished to compute TCs" << endl;



#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
    s.record_remote_stats_in_lsb();
    s.zero_all_stats();
#endif
  }
#endif

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n_local; ++i) {
    tc_vals[i] += local_tc_values[i];
    tc_vals[i] = tc_vals[i]/2;
  }

  int64_t* tc_vals_copy = new int64_t[n_local]();
  memcpy(&tc_vals_copy[0], &tc_vals[0], sizeof(int64_t) * n_local);

  delete [] local_tc_values;
  foMPI_Win_free(&tc_vals_win);

  foMPI_Win_free(&adj_local_win);
  foMPI_Win_free(&offsets_local_win);

  *tc_vals_out = tc_vals_copy;

  if(G->rank_ == 0) {
    for(int i = 0; i < 25; ++i) {
      cout << (*tc_vals_out)[i] << " ";
    }
    cout << endl;
  }

  if(G->rank_ == 0) cout << "Finishing..." << endl;
}





























void TC::runPull_dist_MPIRMA(TradArrayDistGraphR* G, int64_t** tc_vals_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
  s.zero_all_stats();
#endif

  //cout << "...a" << endl;

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Set_Rparam_int("p", G->rank_);
    LSB_Set_Rparam_int("itr", 0);
    LSB_Set_Rparam_string("phs", "prep");
    LSB_Res();
  }
#endif

  //cout << "...b" << endl;
  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* local_tc_values = new int64_t[n_local](); 
  //int64_t* tc_vals = NULL; 

  //foMPI_Win tc_vals_win;
  //foMPI_Win_allocate(n_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &tc_vals, &tc_vals_win);

  //cout << "...c" << endl;
  foMPI_Win offsets_local_win;
  foMPI_Win adj_local_win;
  foMPI_Win_create(G->offsets_local_, (n_local+1)*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &offsets_local_win);
  //cout << "...d" << endl;
  foMPI_Win_create(G->adj_local_, m_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &adj_local_win);
  //cout << "...e" << endl;

  //foMPI_Win_fence(0, tc_vals_win);
  foMPI_Win_fence(0, adj_local_win);

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);

    LSB_Set_Rparam_string("phs", "comp");
    LSB_Res();
  }
#endif

  v_id* offsets_local = G->offsets_local_;
  v_id* adj_local = G->adj_local_;

  v_id v_l_offset = -1;
  v_id v_l_deg = -1;
  v_id neigh_id_g = -1;
  v_id neigh_offset = -1;
  v_id neigh_offset_next = -1;
  v_id neigh_deg = -1;

  v_id neigh_id_l = -1;
  int neigh_owner = -1;

  //cout << "...f" << endl;
  for (v_id v_l = 0; v_l < n_local; ++v_l) {
    v_l_offset = offsets_local[v_l];
    v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

    for (v_id i = 0; i < v_l_deg; ++i) {
      neigh_id_g = adj_local[v_l_offset + i];
      neigh_id_l = G->getVertexLocalID(neigh_id_g);
      neigh_owner = G->getVertexOwningRank(neigh_id_g);

      if(neigh_owner == G->rank_) {
        neigh_offset = offsets_local[neigh_id_l];
        neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

        for (v_id it = 0; it < v_l_deg; ++it) {
          for (v_id e = 0; e < neigh_deg; ++e) {
            if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);

              ++(local_tc_values[v_l]);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }
      }
      else {
        int64_t data[2];
        foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
        foMPI_Win_flush(neigh_owner, offsets_local_win);

        neigh_offset = data[0];
        neigh_offset_next = data[1];
        neigh_deg = neigh_offset_next - neigh_offset;

        //TODO: does it consume too much memory?
        v_id* neighbors = new v_id[neigh_deg]();

        foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
        foMPI_Win_flush(neigh_owner, adj_local_win);

        v_id adj_local_1 = -1;
        v_id adj_local_2 = -1;
        for (v_id it = 0; it < v_l_deg; ++it) {
          adj_local_1 = adj_local[v_l_offset + it];

          for (v_id e = 0; e < neigh_deg; ++e) {
            adj_local_2 = neighbors[e];

            if (adj_local_1 == adj_local_2){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              ++(local_tc_values[v_l]);
#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }

        delete [] neighbors;
      }
    }
  }

  //foMPI_Win_fence(0, tc_vals_win);
  foMPI_Win_fence(0, adj_local_win);

  if(G->rank_ == 0) cout << "Finished to compute TCs" << endl;

#ifdef __BENCHMARK__

  if(G->rank_ == 0) {
    LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
    s.record_remote_stats_in_lsb();
    s.zero_all_stats();
#endif
  }
#endif

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n_local; ++i) {
    local_tc_values[i] = local_tc_values[i]/2;
  }

  int64_t* tc_vals_copy = new int64_t[n_local]();
  memcpy(&tc_vals_copy[0], &local_tc_values[0], sizeof(int64_t) * n_local);

  *tc_vals_out = tc_vals_copy;

  delete [] local_tc_values;

  foMPI_Win_free(&adj_local_win);
  foMPI_Win_free(&offsets_local_win);

  if(G->rank_ == 0) cout << "Finishing..." << endl;
  if(G->rank_ == 0) {
    for(int i = 0; i < 25; ++i) {
      cout << (*tc_vals_out)[i] << " ";
    }
    cout << endl;
  }
}



void TC::runPush_dist_RMA_MPI_plus_MPI_Atomics(TradArrayDistGraphR* G, int64_t** tc_vals_out) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
  s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Set_Rparam_int("p", G->rank_);
    LSB_Set_Rparam_int("itr", 0);
    LSB_Set_Rparam_string("phs", "prep");
    LSB_Res();
  }
#endif

  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* tc_vals = NULL; 

  foMPI_Win tc_vals_win;
  foMPI_Win_allocate(n_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &tc_vals, &tc_vals_win);

  foMPI_Win offsets_local_win;
  foMPI_Win adj_local_win;
  foMPI_Win_create(G->offsets_local_, (n_local+1)*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &offsets_local_win);
  foMPI_Win_create(G->adj_local_, m_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &adj_local_win);

  int64_t* local_tc_values = new int64_t[n_local]();

  foMPI_Win_fence(0, tc_vals_win);

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);
    LSB_Set_Rparam_string("phs", "comp");

    LSB_Res();
  }
#endif

  v_id* offsets_local = G->offsets_local_;
  v_id* adj_local = G->adj_local_;

  v_id v_l_offset = -1;
  v_id v_l_deg = -1;
  v_id neigh_id_g = -1;
  v_id neigh_offset = -1;
  v_id neigh_offset_next = -1;
  v_id neigh_deg = -1;

  v_id neigh_id_l = -1;
  int neigh_owner = -1;

  MPI_Barrier(G->comm_);

  if(G->rank_ == 0) cout << "Start to compute TCs" << endl;

  for (v_id v_l = 0; v_l < n_local; ++v_l) {
    v_l_offset = offsets_local[v_l];
    v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

    for (v_id i = 0; i < v_l_deg; ++i) {
      neigh_id_g = adj_local[v_l_offset + i];
      neigh_id_l = G->getVertexLocalID(neigh_id_g);
      neigh_owner = G->getVertexOwningRank(neigh_id_g);

      if(neigh_owner == G->rank_) {
        neigh_offset = offsets_local[neigh_id_l];
        neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

        for (v_id it = 0; it < v_l_deg; ++it) {
          for (v_id e = 0; e < neigh_deg; ++e) {
            if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              ++(local_tc_values[neigh_id_l]);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }
      }
      else {
        int64_t data[2];
        foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
        foMPI_Win_flush(neigh_owner, offsets_local_win);

        neigh_offset = data[0];
        neigh_offset_next = data[1];
        neigh_deg = neigh_offset_next - neigh_offset;

        //TODO: does it consume too much memory?
        v_id* neighbors = new v_id[neigh_deg]();

        foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
        foMPI_Win_flush(neigh_owner, adj_local_win);

        v_id adj_local_1 = -1;
        v_id adj_local_2 = -1;
        for (v_id it = 0; it < v_l_deg; ++it) {
          adj_local_1 = adj_local[v_l_offset + it];

          for (v_id e = 0; e < neigh_deg; ++e) {
            adj_local_2 = neighbors[e];

            if (adj_local_1 == adj_local_2){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              int64_t add = 1;
              int64_t res = 0;
              v_id local_id = G->getLocalVertexDegree(adj_local_2);
              foMPI_Get_accumulate(&add, 1, MPI_INT64_T, &res, 1, MPI_INT64_T, neigh_owner, neigh_id_l, 1, MPI_INT64_T, foMPI_SUM, tc_vals_win);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.remote_fompi_gac_sum_;
  }
#endif
            }
          }
        }
        delete [] neighbors;
      }
    }
  }

  foMPI_Win_fence(0, tc_vals_win);

  if(G->rank_ == 0) cout << "Finished to compute TCs" << endl;



#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
    s.record_remote_stats_in_lsb();
    s.zero_all_stats();
#endif
  }
#endif

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n_local; ++i) {
    tc_vals[i] += local_tc_values[i];
    tc_vals[i] = tc_vals[i]/2;
  }

  int64_t* tc_vals_copy = new int64_t[n_local]();
  memcpy(&tc_vals_copy[0], &tc_vals[0], sizeof(int64_t) * n_local);

  delete [] local_tc_values;
  foMPI_Win_free(&tc_vals_win);

  foMPI_Win_free(&adj_local_win);
  foMPI_Win_free(&offsets_local_win);

  *tc_vals_out = tc_vals_copy;

  if(G->rank_ == 0) {
    for(int i = 0; i < 25; ++i) {
      cout << (*tc_vals_out)[i] << " ";
    }
    cout << endl;
  }

  if(G->rank_ == 0) cout << "Finishing..." << endl;
}

/*
void TC::runPush_dist_MP(TradArrayDistGraphR* G, int64_t** tc_vals_out) {

#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s;
  s.zero_all_stats();
#endif

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Set_Rparam_int("p", G->rank_);
    LSB_Set_Rparam_int("itr", 0);
    LSB_Set_Rparam_string("phs", "prep");
    LSB_Res();
  }
#endif

  v_id n = G->n_;
  v_id n_local = G->n_local_;
  v_id m_local = G->m_local_;

  int64_t* tc_vals = NULL; 

  foMPI_Win tc_vals_win;
  foMPI_Win_allocate(n_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &tc_vals, &tc_vals_win);

  foMPI_Win offsets_local_win;
  foMPI_Win adj_local_win;
  foMPI_Win_create(G->offsets_local_, (n_local+1)*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &offsets_local_win);
  foMPI_Win_create(G->adj_local_, m_local*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, G->comm_, &adj_local_win);

  int64_t* local_tc_values = new int64_t[n_local]();

  foMPI_Win_fence(0, tc_vals_win);

#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);
    LSB_Set_Rparam_string("phs", "comp");

    LSB_Res();
  }
#endif

  v_id* offsets_local = G->offsets_local_;
  v_id* adj_local = G->adj_local_;

  v_id v_l_offset = -1;
  v_id v_l_deg = -1;
  v_id neigh_id_g = -1;
  v_id neigh_offset = -1;
  v_id neigh_offset_next = -1;
  v_id neigh_deg = -1;

  v_id neigh_id_l = -1;
  int neigh_owner = -1;

  int* finished = new int[G->num_procs_]();
  finished[G->rank_] = 1;

  vector<int64_t>** out_buf = new vector<int64_t>*[G->num_procs_]();

  for(int i = 0; i < G->num_procs_; ++i) {
    if(i == G->rank_) {
      continue;
    }

    out_buf[i] = new vector<int64_t>();
  }

  MPI_Barrier(G->comm_);

  if(G->rank_ == 0) cout << "Start to compute TCs" << endl;

  for (v_id v_l = 0; v_l < n_local; ++v_l) {
    v_l_offset = offsets_local[v_l];
    v_l_deg = offsets_local[v_l + 1] - offsets_local[v_l];

    for (v_id i = 0; i < v_l_deg; ++i) {
      neigh_id_g = adj_local[v_l_offset + i];
      neigh_id_l = G->getVertexLocalID(neigh_id_g);
      neigh_owner = G->getVertexOwningRank(neigh_id_g);

      if(neigh_owner == G->rank_) {
        neigh_offset = offsets_local[neigh_id_l];
        neigh_deg = offsets_local[neigh_id_l + 1] - offsets_local[neigh_id_l];

        for (v_id it = 0; it < v_l_deg; ++it) {
          for (v_id e = 0; e < neigh_deg; ++e) {
            if (adj_local[v_l_offset + it] == adj_local[neigh_offset + e]){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              ++(local_tc_values[neigh_id_l]);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.local_fompi_gac_sum_;
  }
#endif
            }
          }
        }
      }
      else {
        int64_t data[2];
        foMPI_Get(&data[0], 2, MPI_INT64_T, neigh_owner, neigh_id_l, 2, MPI_INT64_T, offsets_local_win);
        foMPI_Win_flush(neigh_owner, offsets_local_win);

        neigh_offset = data[0];
        neigh_offset_next = data[1];
        neigh_deg = neigh_offset_next - neigh_offset;

        //TODO: does it consume too much memory?
        v_id* neighbors = new v_id[neigh_deg]();

        foMPI_Get(neighbors, neigh_deg, MPI_INT64_T, neigh_owner, neigh_offset, neigh_deg, MPI_INT64_T, adj_local_win);
        foMPI_Win_flush(neigh_owner, adj_local_win);

        v_id adj_local_1 = -1;
        v_id adj_local_2 = -1;
        for (v_id it = 0; it < v_l_deg; ++it) {
          adj_local_1 = adj_local[v_l_offset + it];

          for (v_id e = 0; e < neigh_deg; ++e) {
            adj_local_2 = neighbors[e];

            if (adj_local_1 == adj_local_2){
              //__sync_fetch_and_add(&tc_vals[adj_local[v_l_offset + i]], 1);
              int64_t add = 1;
              int64_t res = 0;
              v_id local_id = G->getLocalVertexDegree(adj_local_2);
            
              out_buf[neigh_owner]->push_back(neigh_id_l);     
              if(out_buf[neigh_owner]->size() == out_buf_max_size_) {
                MPI_Send(out_buf[neigh_owner]->data(), out_buf[neigh_owner]->size(), MPI_INT64_T, neigh_owner, 0, G->comm_);
              }
            
//              foMPI_Get_accumulate(&add, 1, MPI_INT64_T, &res, 1, MPI_INT64_T, neigh_owner, neigh_id_l, 1, MPI_INT64_T, foMPI_SUM, tc_vals_win);

#ifdef __XGRAPH_COUNT_OPERATIONS__
  if(G->rank_ == 0) {
              ++s.remote_fompi_gac_sum_;
  }
#endif
            }
          }
        }
        delete [] neighbors;
      }
    }
  }

  for(int i = 0; i < G->num_procs_; ++i) {
    if(i == G->rank_) {
      continue;
    }

    MPI_Send(out_buf[i]->data(), out_buf[i]->size(), MPI_INT64_T, i, 0, G->comm_);

    int all_went = 1;
    MPI_Send(&all_went, 1, MPI_INT, i, 1, G->comm_);
  }

  int done_procs = 1;
  int flag;
  MPI_Status s;
  while(done_procs < G->num_procs_) {
    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, G->comm_, &flag, &s);
    if(flag) {
      if(s.MPI_TAG == 0) {
        int num = 0;
        MPI_Get_count(&s, MPI_INT64_T, &num);
        assert(num > 0);

        int64_t* vals = new int64_t[num];
        MPI_Recv(vals, num, MPI_INT64_T, s.MPI_SOURCE, s.MPI_TAG, G->comm_, &s);

        for(int x = 0; x < num; ++x) {
              ++(local_tc_values[ vals[x] ]);
        }

        delete [] vals;

      }
      else if(s.MPI_TAG == 1) {
        ++done_procs;
        finished[s.MPI_SOURCE] = 1;
      }
      else {
        assert(false);
      }
    }
  }

  foMPI_Win_fence(0, tc_vals_win);

  if(G->rank_ == 0) cout << "Finished to compute TCs" << endl;



#ifdef __BENCHMARK__
  if(G->rank_ == 0) {
    LSB_Rec(run_id);

#ifdef __XGRAPH_COUNT_OPERATIONS__
    s.record_remote_stats_in_lsb();
    s.zero_all_stats();
#endif
  }
#endif

  // it is not necessary if we are using sorted tables
  for (unsigned i = 0; i < n_local; ++i) {
    tc_vals[i] += local_tc_values[i];
    tc_vals[i] = tc_vals[i]/2;
  }

  int64_t* tc_vals_copy = new int64_t[n_local]();
  memcpy(&tc_vals_copy[0], &tc_vals[0], sizeof(int64_t) * n_local);

  delete [] local_tc_values;
  foMPI_Win_free(&tc_vals_win);

  foMPI_Win_free(&adj_local_win);
  foMPI_Win_free(&offsets_local_win);

  *tc_vals_out = tc_vals_copy;

  if(G->rank_ == 0) {
    for(int i = 0; i < 25; ++i) {
      cout << (*tc_vals_out)[i] << " ";
    }
    cout << endl;
  }

  if(G->rank_ == 0) cout << "Finishing..." << endl;
}
*/


#endif
